"Copy and paste to somewhere it should be"

from typing import TypedDict

__all__ = ["EventTemplate"]


class EventTemplate(TypedDict):
    """Event template for event sourcing"""

    name: str  # event type, like 'user.login'
    id: str  # event id, such as type + timestamp + random
    prev_id: str  # previous(parent) event id, for event chain
    root_id: str  # origin event id, the first event id in chain
    version: str  # event version, optional
    creator: str  # event creator, client name or ip or user name
    create_at: int  # event create time, secs or ms since epoch
    expire_at: int  # event expire time, 0 for never expire
    payload: dict  # event payload data



import asyncio
import atexit
import base64
import collections.abc
import gzip
import hashlib
import inspect
import json
import os
import pickle
import re
import subprocess
import sys
import tempfile
import timeit
import traceback
import types
from collections import UserDict
from datetime import datetime
from enum import IntEnum
from functools import wraps
from itertools import groupby, islice
from os.path import basename, exists
from pathlib import Path
from threading import Lock, Timer, current_thread
from time import gmtime, mktime, strftime, strptime, time, timezone
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

__all__ = [
    "ttime",
    "ptime",
    "slice_into_pieces",
    "slice_by_size",
    "unique",
    "retry",
    "guess_interval",
    "get_hash",
    "find_jsons",
    "code_inline",
    "read_size",
    "read_time",
    "Validator",
    "stagger_sort",
    "default_dict",
    "always_return_value",
    "format_error",
    "Trie",
    "GuessExt",
    "xor_encode_decode",
    "is_running",
    "set_pid_file",
    "get_paste",
    "set_clip",
    "switch_flush_print",
    "unix_rlimit",
    "SimpleFilter",
    "FileDict",
    "PathLock",
    "i2b",
    "b2i",
    "get_hash_int",
    "iter_weights",
    "get_size",
    "base_encode",
    "base_decode",
    "gen_id",
    "timeti",
    "SnowFlake",
    "cut_file",
]


def ttime(
    timestamp: Optional[Union[float, int]] = None,
    tzone: int = int(-timezone / 3600),
    fmt="%Y-%m-%d %H:%M:%S",
) -> str:
    """Converts a timestamp to a human-readable timestring formatted as %Y-%m-%d %H:%M:%S.

    >>> ttime(1486572818.421858323, tzone=8)
    '2017-02-09 00:53:38'

    Args:
        timestamp (float, optional): the timestamp float. Defaults to time.time().
        tzone (int, optional): time compensation. Defaults to int(-time.timezone / 3600).
        fmt (str, optional): strftime fmt. Defaults to "%Y-%m-%d %H:%M:%S".

    Returns:
        str: time string formatted.
    """
    fix_tz = tzone * 3600
    timestamp = time() if timestamp is None else timestamp
    return strftime(fmt, gmtime(timestamp + fix_tz))


def ptime(
    timestring: Optional[str] = None,
    tzone: int = int(-timezone / 3600),
    fmt: str = "%Y-%m-%d %H:%M:%S",
) -> int:
    """Converts a timestring formatted as %Y-%m-%d %H:%M:%S back into a timestamp.

    >>> ptime("2018-03-15 01:27:56", tzone=8)
    1521048476

    Args:
        timestring (str, optional): string like 2018-03-15 01:27:56. Defaults to ttime().
        tzone (int, optional): time compensation. Defaults to int(-timezone / 3600).
        fmt (_type_, optional): strptime fmt. Defaults to "%Y-%m-%d %H:%M:%S".

    Returns:
        str: time string formatted.
    """
    fix_tz = -(tzone * 3600 + timezone)
    #: str(timestring) for datetime.datetime object
    if timestring:
        return int(mktime(strptime(str(timestring), fmt)) + fix_tz)
    else:
        return int(time())


def slice_into_pieces(items: Sequence, n: int) -> Generator[tuple, None, None]:
    """Divides a sequence into “n” segments, returning a generator that yields “n” pieces.

    >>> for chunk in slice_into_pieces(range(10), 3):
    ...     print(chunk)
    (0, 1, 2, 3)
    (4, 5, 6, 7)
    (8, 9)

    Args:
        seq (_type_): input a sequence.
        n (_type_): split the given sequence into "n" pieces.

    Returns:
        Generator[tuple, None, None]: a generator with tuples.

    Yields:
        Iterator[Generator[tuple, None, None]]: a tuple with n of items.
    """
    length = len(items)
    if length % n == 0:
        size = length // n
    else:
        size = length // n + 1
    for it in slice_by_size(items, size):
        yield it


def slice_by_size(
    items: Sequence, size: int, callback=tuple
) -> Generator[tuple, None, None]:
    """Slices a sequence into chunks of a specified “size”, returning a generator that produces tuples of chunks.

    >>> for chunk in slice_by_size(range(10), 3):
    ...     print(chunk)
    (0, 1, 2)
    (3, 4, 5)
    (6, 7, 8)
    (9,)

    Args:
        items (Sequence): _description_
        size (int): _description_

    Returns:
        Generator[tuple, None, None]: a generator with tuples.

    Yields:
        Iterator[Generator[tuple, None, None]]: a tuple with n of items.
    """
    iter_seq = iter(items)
    while True:
        chunk = callback(islice(iter_seq, size))
        if chunk:
            yield chunk
        else:
            break


def unique(
    items: Sequence, key: Optional[Callable] = None
) -> Generator[Any, None, None]:
    """Removes duplicate elements from a sequence while preserving the original order efficiently.

    >>> a = ['01', '1', '2']
    >>> list(unique(a, int))
    [1, 2]
    >>> list(unique(a))
    ['01', '1', '2']

    Args:
        items (Sequence): raw sequence.
        key (Callable): the function to normalize each item.

    Returns:
        Generator[tuple, None, None]: a generator with unique items.

    Yields:
        Iterator[Generator[tuple, None, None]]: the unique item.
    """
    seen: set[Any] = set()
    _add = seen.add
    if key:
        for item in items:
            _item = key(item)
            if _item not in seen:
                yield _item
                _add(_item)
    else:
        for item in items:
            if item not in seen:
                yield item
                _add(item)


def retry(
    tries=2,
    exceptions: Tuple[Type[BaseException]] = (Exception,),
    return_exception=False,
):
    """A decorator that retries the decorated function up to “tries” times if the specified exceptions are raised.

    >>> func = lambda items: 1/items.pop(0)
    >>> items = [0, 1]
    >>> new_func = retry(tries=2, exceptions=(ZeroDivisionError,))(func)
    >>> new_func(items)
    1.0

    Args:
        tries (int, optional): try n times, if n==1 means no retry. Defaults to 1.
        exceptions (Tuple[Type[BaseException]], optional): only retry the given errors. Defaults to (Exception,).
        return_exception (bool, optional): raise the last exception or return it. Defaults to False.
    """

    def wrapper(function):
        @wraps(function)
        def retry_sync(*args, **kwargs):
            error = None
            for _ in range(tries):
                try:
                    return function(*args, **kwargs)
                except exceptions as err:
                    error = err
            if return_exception:
                return error
            raise error

        @wraps(function)
        async def retry_async(*args, **kwargs):
            error = None
            for _ in range(tries):
                try:
                    return await function(*args, **kwargs)
                except exceptions as err:
                    error = err
            if return_exception:
                return error
            raise error

        if asyncio.iscoroutinefunction(function):
            return retry_async
        else:
            return retry_sync

    return wrapper


def guess_interval(nums, accuracy=0):
    """Analyzes a sequence of numbers and returns the median, calculating intervals only if they are greater than or equal to the specified accuracy.

    >>> # sorted_seq: [2, 10, 12, 19, 19, 29, 30, 32, 38, 40, 41, 54, 62]
    >>> # diffs: [8, 7, 10, 6, 13, 8]
    >>> # median: 8
    >>> seq = [2, 10, 12, 19, 19, 29, 30, 32, 38, 40, 41, 54, 62]
    >>> guess_interval(seq, 5)
    8

    """
    if not nums:
        return 0
    nums = sorted([int(i) for i in nums])
    if len(nums) == 1:
        return nums[0]
    diffs = [nums[i + 1] - nums[i] for i in range(len(nums) - 1)]
    diffs = [item for item in diffs if item >= accuracy]
    sorted_diff = sorted(diffs)
    return sorted_diff[len(diffs) // 2]


def get_hash(
    string,
    n: Optional[Union[Tuple[int, int], List[int], int]] = None,
    default: Callable = lambda obj: str(obj).encode("utf-8"),
    func=hashlib.md5,
) -> str:
    """Generates a hash string from the given input string.

    >>> get_hash(123456, 10)
    'a59abbe56e'
    >>> get_hash('test')
    '098f6bcd4621d373cade4e832627b4f6'
    >>> get_hash(['list_demo'], (5, 10))
    '7152a'
    >>> get_hash(['list_demo'], func=hashlib.sha256)
    'a6072e063d36a09052a9e5eb389a425a3dc158d3a5955808159a118aa192c718'
    >>> get_hash(['list_demo'], 16, func=hashlib.sha256)
    '389a425a3dc158d3'
    """
    if isinstance(string, bytes):
        _bytes = string
    else:
        _bytes = default(string)
    _temp = func(_bytes).hexdigest()
    if n is None:
        return _temp
    elif isinstance(n, (int, float)):
        start, end = (len(_temp) - n) // 2, (n - len(_temp)) // 2
        return _temp[start:end]
    elif isinstance(n, (tuple, list)):
        start, end = n[0], n[1]
        return _temp[start:end]


def get_hash_int(
    string,
    n: int = 16,
    default: Callable = lambda obj: str(obj).encode("utf-8"),
    func=hashlib.md5,
) -> int:
    """Generates a int hash(like docid) from the given input bytes.

    >>> get_hash_int(1)
    2035485573088411
    >>> get_hash_int("string")
    1418352543534881
    >>> get_hash_int(b'123456', 16)
    4524183350839358
    >>> get_hash_int(b'123', 10)
    5024125808
    >>> get_hash_int(b'123', 13, func=hashlib.sha256)
    1787542395619
    >>> get_hash_int(b'123', 13, func=hashlib.sha512)
    3045057537218
    >>> get_hash_int(b'123', 13, func=hashlib.sha1)
    5537183137519
    """
    return int(get_hash(string, n=None, default=default, func=func), 16) % 10**n


def find_jsons(
    string,
    return_as: Literal["json", "object", "index"] = "json",
    json_loader=json.loads,
):
    """A generator that locates valid JSON strings, supporting only dictionaries and lists.

    >>> list(find_jsons('string["123"]123{"a": 1}[{"a": 1, "b": [1,2,3]}]'))
    ['["123"]', '{"a": 1}', '[{"a": 1, "b": [1,2,3]}]']
    >>> list(find_jsons('string[]{}{"a": 1}'))
    ['[]', '{}', '{"a": 1}']
    >>> list(find_jsons('string[]|{}string{"a": 1}', return_as='index'))
    [(6, 8), (9, 11), (17, 25)]
    >>> list(find_jsons('xxxx[{"a": 1, "b": [1,2,3]}]xxxx', return_as='object'))
    [[{'a': 1, 'b': [1, 2, 3]}]]
    """

    def find_matched(string, left, right):
        _stack = []
        for index, char in enumerate(string):
            if char == left:
                _stack.append(index)
            elif char == right:
                try:
                    _stack.pop()
                except IndexError:
                    break
            else:
                continue
            if not _stack:
                return index

    search = re.search
    brackets_map = {"{": "}", "[": "]"}
    current_start = 0
    while string and isinstance(string, str):
        _match = search(r"[\[\{]", string)
        if not _match:
            break
        left = _match.group()
        right = brackets_map[left]
        _start = _match.span()[0]
        sub_string = string[_start:]
        _end = find_matched(sub_string, left, right)
        if _end is None:
            # not found matched, check next left
            string = sub_string
            continue
        string = sub_string[_end + 1 :]
        try:
            _partial = sub_string[: _end + 1]
            _loaded_result = json_loader(_partial)
            yield {
                "json": _partial,
                "object": _loaded_result,
                "index": (current_start + _start, current_start + _start + _end + 1),
            }.get(return_as, string)
        except (ValueError, TypeError):
            pass
        current_start += _start + _end + 1


def code_inline(
    source_code: str,
    encoder: Literal["b16", "b32", "b64", "b85"] = "b85",
) -> str:
    """Minifies Python source code into a single line.

    >>> code1 = code_inline('def test_code1(): return 12345')
    >>> code1
    'import base64,gzip;exec(gzip.decompress(base64.b85decode("ABzY80RR910{=@%O;adIEiQ>q&QD1-)X=n2C`v6UEy`0cG%_|Z1psqiSP>oo000".encode("u8"))))'
    >>> exec(code1)
    >>> test_code1()
    12345
    >>> code2 = code_inline("v=12345")
    >>> code2
    'import base64,gzip;exec(gzip.decompress(base64.b85decode("ABzY80RR910{<(sH8e6dF$Dk}<L9Rb0000".encode("u8"))))'
    >>> exec(code2)
    >>> v
    12345

    Args:
        source_code (str): python original code.
        encoder (Literal['b16', 'b32', 'b64', 'b85'], optional): base64.encoder. Defaults to "b85".
    Returns:
        new source code inline.
    """
    _encoder = getattr(base64, f"{encoder}encode")
    _source = source_code.encode(encoding="u8")
    _source = gzip.compress(_source, mtime=1)
    _source = _encoder(_source)
    code = _source.decode("u8")
    result = (
        "import base64,gzip;exec(gzip.decompress("
        f'base64.{encoder}decode("{code}".encode("u8"))))'
    )
    return result


class BytesUnit(IntEnum):
    B = 0
    KB = 1 * 1024**1
    MB = 1 * 1024**2
    GB = 1 * 1024**3
    TB = 1 * 1024**4
    PB = 1 * 1024**5
    EB = 1 * 1024**6
    ZB = 1 * 1024**7
    YB = 1 * 1024**8


class TimeUnit(IntEnum):
    secs = 0
    mins = 60
    hours = 60 * 60
    days = 60 * 60 * 24
    mons = 60 * 60 * 24 * 30
    years = 60 * 60 * 24 * 365


def read_num(
    b,
    enum: Type[IntEnum] = IntEnum,
    rounded: Optional[int] = None,
    shorten=False,
    precision=1.0,
    sep=" ",
):
    units = tuple(enum)
    unit = units[0]
    ntime = 0.5 if rounded and rounded > 0 else 1
    for _unit in units:
        if b >= _unit.value * ntime:
            unit = _unit
        else:
            break
    result = round(b / (unit.value or 1), rounded)
    if shorten and isinstance(result, float):
        int_result = int(result)
        if result:
            if int_result / result >= precision:
                result = int_result
    return f"{result}{sep}{unit.name}"


def read_size(b, rounded: Optional[int] = None, shorten=False, precision=1.0, sep=" "):
    """Converts byte counts into a human-readable string. Setting shorten=True and precision=0.99 will trim unnecessary decimal places from the tail of floating-point numbers.

    >>> (read_size(1023), read_size(1024))
    ('1023 B', '1 KB')
    >>> (read_size(400.5, 1), read_size(400.5, 1, True), read_size(400.5, 1, True, 0.99))
    ('400.5 B', '400.5 B', '400 B')
    >>> (read_size(511.55, 1, shorten=True, precision=0.999), read_size(512.55, 1, shorten=True, precision=0.999))
    ('511.6 B', '0.5 KB')
    >>> (read_size(511, 1, shorten=True), read_size(512, 1, shorten=True))
    ('511 B', '0.5 KB')
    >>> read_size(512, 1, sep='')
    '0.5KB'
    >>> read_size(1025, 1, shorten=False), read_size(1025, 1, shorten=True)
    ('1.0 KB', '1 KB')
    >>> for i in range(0, 5):
    ...     [1.1111 * 1024**i, i, read_size(1.1111 * 1024**i, rounded=i)]
    ...
    [1.1111, 0, '1.0 B']
    [1137.7664, 1, '1.1 KB']
    [1165072.7936, 2, '1.11 MB']
    [1193034540.6464, 3, '1.111 GB']
    [1221667369621.9136, 4, '1.1111 TB']

    Args:
        b: bytes
        rounded (int, optional): arg for round. Defaults to None.
        shorten (bool): shorten unnecessary tail 0.
        precision: (float): shorten precision, often set to 0.99.
        sep (str, optional): sep between result and unit.

    Returns:
        str


    """
    return read_num(
        b,
        BytesUnit,
        rounded=rounded,
        shorten=shorten,
        precision=precision,
        sep=sep,
    )


def read_time(
    secs, rounded: Optional[int] = None, shorten=False, precision=1.0, sep=" "
):
    """Converts seconds into a more readable time duration string.

    >>> read_time(0)
    '0 secs'
    >>> read_time(60)
    '1 mins'
    >>> for i in range(0, 6):
    ...     [1.2345 * 60**i, read_time(1.2345 * 60**i, rounded=1)]
    ...
    [1.2345, '1.2 secs']
    [74.07, '1.2 mins']
    [4444.2, '1.2 hours']
    [266652.0, '3.1 days']
    [15999120.0, '0.5 years']
    [959947200.0, '30.4 years']

    Args:
        b: seconds
        rounded (int, optional): arg for round. Defaults to None.
        shorten (bool): shorten unnecessary tail 0.
        precision: (float): shorten precision, often set to 0.99.
        sep (str, optional): sep between result and unit.

    Returns:
        str
    """
    return read_num(
        secs,
        TimeUnit,
        rounded=rounded,
        shorten=shorten,
        precision=precision,
        sep=sep,
    )


class Validator:
    """Validator for dataclasses.
    >>> from dataclasses import dataclass, field
    >>>
    >>>
    >>> @dataclass
    ... class Person(Validator):
    ...     screen: dict = field(metadata={"callback": lambda i: i.clear() or {'s': 4}})
    ...     name: str = field(default=None, metadata={"callback": str})
    ...     age: int = field(default=0, metadata={"callback": int})
    ...     other: str = ''
    ...
    >>>
    >>> # test type callback
    >>> print(Person({"s": 3}, 1, "1"))
    Person(screen={'s': 4}, name='1', age=1, other='')

    >>> # test Validator.STRICT = False, `other` could be int
    >>> Validator.STRICT = False
    >>> Person({"s": 3}, "1", "1", 0)
    Person(screen={'s': 4}, name='1', age=1, other=0)

    >>> # test Validator.STRICT = True, raise TypeError, `other` should not be int
    >>> Validator.STRICT = True
    >>> try:
    ...     Person({"s": 3}, "1", "1", 0)
    ... except TypeError as e:
    ...     print(e)
    `other` should be `str` but given `int`
    >>> # test class Name
    >>> @dataclass
    ... class Name(Validator):
    ...     name: str
    ...
    >>> @dataclass
    ... class Person(Validator):
    ...     name: Name
    ...
    >>> try:
    ...     print(Person('name'))
    ... except TypeError as e:
    ...     print(e)
    ...
    `name` should be `Name` but given `str`
    >>> # test typing.Dict[str, str]
    >>> import typing
    >>> @dataclass
    ... class Person(Validator):
    ...     name: typing.Dict[str, str]
    ...
    >>> try:
    ...     print(Person('name'))
    ... except TypeError as e:
    ...     print(e)
    ...
    `name` should be `dict` but given `str`
    """

    STRICT = True

    def __post_init__(self):
        for f in self.__dataclass_fields__.values():
            callback = f.metadata.get("callback")
            value = getattr(self, f.name)
            if callback:
                setattr(self, f.name, callback(value))
            if self.STRICT:
                value = getattr(self, f.name)
                if inspect.isclass(f.type):
                    tp = f.type
                elif hasattr(f.type, "__origin__"):
                    tp = f.type.__origin__
                else:
                    continue
                if not isinstance(value, tp):
                    raise TypeError(
                        f"`{f.name}` should be `{tp.__name__}` but given `{type(value).__name__}`"
                    )

    def quick_to_dict(self):
        "not very reliable"
        return {
            field.name: getattr(self, field.name)
            for field in self.__dataclass_fields__.values()
        }


def stagger_sort(items: Sequence, group_key, sort_key=None, sort_reverse=False):
    """Ensures that identical groups are ordered and evenly distributed, mitigating data skew. The function does not alter the original list and returns a generator.

    >>> items = [('a', 0), ('a', 2), ('a', 1), ('b', 0), ('b', 1)]
    >>> list(stagger_sort(items, sort_key=lambda i: (i[0], i[1]), group_key=lambda i: i[0]))
    [('a', 0), ('b', 0), ('a', 1), ('b', 1), ('a', 2)]
    >>> items = ['a-a', 'a-b', 'b-b', 'b-c', 'b-a', 'b-d', 'c-a', 'c-a']
    >>> list(stagger_sort(items, sort_key=lambda i: (i[0], i[2]), group_key=lambda i: i[0]))
    ['a-a', 'b-a', 'c-a', 'a-b', 'b-b', 'c-a', 'b-c', 'b-d']
    """
    if sort_key:
        items = sorted(items, key=sort_key, reverse=sort_reverse)
    else:
        items = list(items)
    buckets = [list(group[1]) for group in groupby(items, group_key)]
    while True:
        next_buckets = []
        for items in buckets:
            try:
                yield items.pop(0)
                next_buckets.append(items)
            except IndexError:
                pass
        if next_buckets:
            buckets = next_buckets
        else:
            break


def iter_weights(
    weight_dict: Dict[Any, Union[int, float]], loop_length=100, round_int=round
):
    """Generates an element sequence based on weights.

    This function produces a sequence of elements where each element's frequency of occurrence
    is proportional to its weight from the provided dictionary. Elements with higher weights
    appear more frequently in the sequence. The total cycle length can be adjusted via the
    `loop_length` parameter. The `round_int` parameter allows customization of the rounding
    function to control the precision of weight calculations.

    Keys with weights greater than 0 will be yielded.

    Parameters:
    - weight_dict: A dictionary where keys are elements and values are their respective weights.
    - loop_length: An integer defining the total length of the repeating cycle, defaulting to 100.
    - round_int: A function used for rounding, defaulting to Python's built-in `round`.

    Yields:
    A generator that yields a sequence of elements distributed according to their weights.

    Examples:
    >>> list(iter_weights({"6": 6, "3": 3, "1": 0.4}, 10))
    ['6', '3', '6', '3', '6', '3', '6', '6', '6']
    >>> list(iter_weights({"6": 6, "3": 3, "1": 0.9}, 10))
    ['6', '3', '1', '6', '3', '6', '3', '6', '6', '6']
    >>> list(iter_weights({"6": 6, "3": 3, "1": 0.9}, 10, round_int=int))
    ['6', '3', '6', '3', '6', '3', '6', '6', '6']
    >>> from itertools import cycle
    >>> c = cycle(iter_weights({"6": 6, "3": 3, "1": 1}, loop_length=10))
    >>> [next(c) for i in range(20)]
    ['6', '3', '1', '6', '3', '6', '3', '6', '6', '6', '6', '3', '1', '6', '3', '6', '3', '6', '6', '6']
    """
    keys = list(weight_dict.keys())
    items = []
    total = sum(weight_dict.values())
    for key, value in weight_dict.items():
        length = round_int(loop_length * value / total) if value > 0 else 0
        items.extend([key] * length)
    for item in stagger_sort(
        items,
        group_key=lambda x: keys.index(x),
        sort_key=lambda x: weight_dict[x],
        sort_reverse=True,
    ):
        yield item


def default_dict(typeddict_class: Type[dict], **kwargs):
    """Initializes a dictionary with default zero values based on a subclass of TypedDict.

    >>> class Demo(dict):
    ...     int_obj: int
    ...     float_obj: float
    ...     bytes_obj: bytes
    ...     str_obj: str
    ...     list_obj: list
    ...     tuple_obj: tuple
    ...     set_obj: set
    ...     dict_obj: dict
    >>> item = default_dict(Demo, bytes_obj=b'1')
    >>> item
    {'int_obj': 0, 'float_obj': 0.0, 'bytes_obj': b'1', 'str_obj': '', 'list_obj': [], 'tuple_obj': (), 'set_obj': set(), 'dict_obj': {}}
    >>> type(item)
    <class 'morebuiltins.utils.Demo'>
    >>> from typing import TypedDict
    >>> class Demo(TypedDict):
    ...     int_obj: int
    ...     float_obj: float
    ...     bytes_obj: bytes
    ...     str_obj: str
    ...     list_obj: list
    ...     tuple_obj: tuple
    ...     set_obj: set
    ...     dict_obj: dict
    >>> item = default_dict(Demo, bytes_obj=b'1')
    >>> item
    {'int_obj': 0, 'float_obj': 0.0, 'bytes_obj': b'1', 'str_obj': '', 'list_obj': [], 'tuple_obj': (), 'set_obj': set(), 'dict_obj': {}}
    >>> type(item)
    <class 'dict'>
    """
    result = typeddict_class()
    built_in_types = {int, float, bytes, str, list, tuple, set, dict}
    for key, tp in typeddict_class.__annotations__.items():
        if key in kwargs:
            result[key] = kwargs[key]
        elif tp in built_in_types:
            result[key] = tp()
    return result


def always_return_value(value):
    """Got a function always return the given value.
    >>> func = always_return_value(1)
    >>> func(1, 2, 3)
    1
    >>> func(1, 2, c=3)
    1
    """

    def func(*args, **kws):
        return value

    return func


def _tb_filter(tb: traceback.FrameSummary):
    # filt tb with code-line and not excluded -packages
    return tb.line and "-packages" not in tb.filename


def format_error(
    error: BaseException,
    index: Union[int, slice] = slice(-3, None, None),
    filter: Optional[Callable] = _tb_filter,
    template="[{trace_routes}] {error_line} >>> {error.__class__.__name__}({error!s:.100})",
    filename_filter: Tuple[str, str] = ("", ""),
    **kwargs,
) -> str:
    r"""Extracts frame information from an exception, with an option to filter out “-packages” details by default. To shorten your exception message.

    Parameters:

    - `error` (`BaseException`): The exception instance for which the stack trace information is to be extracted and formatted.
    - `index` (`Union[int, slice]`, optional): Specifies which frames to include in the output. By default, it's set to `slice(-3, None, None)`, showing the last three frames. Can be an integer for a single frame or a slice object for a range of frames.
    - `filter` (`Optional[Callable]`, optional): A callable that determines whether a given frame should be included. Defaults to `_tb_filter`, which typically filters out frames from "-packages". If set to `None`, no filtering occurs.
    - `template` (`str`, optional): A string template defining how the error message should be formatted. It can include placeholders like `{trace_routes}`, `{error_line}`, and `{error.__class__.__name__}`. The default template provides a concise summary of the error location and type.
    - `filename_filter` (`Tuple[str, str]`, optional): A tuple specifying the include and exclude strings of filename. Defaults to `("", "")`, which means no filtering occurs.
    - `**kwargs`: Additional keyword arguments to be used within the formatting template.

    Returns:

    A string representing the formatted error message based on the provided parameters and template.

    Demo:

    >>> try:
    ...     # test default
    ...     1 / 0
    ... except Exception as e:
    ...     format_error(e)
    '[<doctest>:<module>(3)] 1 / 0 >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test in function
    ...     def func1(): 1 / 0
    ...     func1()
    ... except Exception as e:
    ...     format_error(e)
    '[<doctest>:<module>(4)|<doctest>:func1(3)] def func1(): 1 / 0 >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test index
    ...     def func2(): 1 / 0
    ...     func2()
    ... except Exception as e:
    ...     format_error(e, index=0)
    '[<doctest>:<module>(4)] func2() >>> ZeroDivisionError(division by zero)'
    >>> try:
    ...     # test slice index
    ...     def func2(): 1 / 0
    ...     func2()
    ... except Exception as e:
    ...     format_error(e, index=slice(-1, None, None))
    '[<doctest>:func2(3)] def func2(): 1 / 0 >>> ZeroDivisionError(division by zero)'
    """
    try:
        if filter:
            _filter = filter
        else:
            _filter = always_return_value(True)
        tbs = [tb for tb in traceback.extract_tb(error.__traceback__) if _filter(tb)]
        if isinstance(index, slice):
            tbs = tbs[index]
        elif isinstance(index, int):
            tbs = [tbs[index]]
        else:
            raise ValueError("Invalid index type")
        trace_route_list = []
        include, exclude = filename_filter
        for tb in tbs:
            filename = tb.filename
            if include and include not in filename:
                continue
            if exclude and exclude in filename:
                continue
            if exists(filename):
                _basename = basename(filename)
            elif filename[0] == "<":
                _basename = f"{filename.split()[0]}>"
            else:
                _basename = filename
            trace_route_list.append(f"{_basename}:{tb.name}({tb.lineno})")
        trace_routes = "|".join(trace_route_list)
        _kwargs = {
            "tbs": tbs,
            "error": error,
            "error_line": tbs[-1].line,
            "trace_routes": trace_routes,
        }
        _kwargs.update(kwargs)
        return template.format_map(_kwargs)
    except IndexError:
        return ""


class Trie(UserDict):
    """Transforms a standard dictionary into a trie structure that supports prefix matching.

    >>> trie = Trie({"ab": 1, "abc": 2, b"aa": 3, ("e", "e"): 4, (1, 2): 5})
    >>> trie
    {'a': {'b': {'_VALUE': 1, 'c': {'_VALUE': 2}}}, 97: {97: {'_VALUE': 3}}, 'e': {'e': {'_VALUE': 4}}, 1: {2: {'_VALUE': 5}}}
    >>> trie.match("abcde")
    (3, 2)
    >>> trie.match("abde")
    (2, 1)
    >>> trie.match(b"aabb")
    (2, 3)
    >>> trie.match("eee")
    (2, 4)
    >>> trie.match(list("eee"))
    (2, 4)
    >>> trie.match(tuple("eee"))
    (2, 4)
    >>> trie.match((1, 2, 3, 4))
    (2, 5)
    """

    def __init__(self, data: Dict[AnyStr, Any], value_node="_VALUE"):
        # data should be 1-depth dict
        super().__init__(self.parse_dict(data, value_node))
        self.value_node = value_node
        self.not_set = object()

    @staticmethod
    def parse_dict(data: dict, value_node="_VALUE") -> dict:
        "Convert a regular dictionary to a prefix tree dictionary"
        result: dict = {}
        for word, value in data.items():
            trie = result
            for char in word:
                if char not in trie:
                    trie[char] = {}
                trie = trie[char]
            trie[value_node] = value
        return result

    def match(self, seq, default: Tuple[Union[None, int], Any] = (None, None)):
        """Use the input sequence to match the hit value in the prefix tree as much as possible.

        Args:
            seq: iterable object
            default:return while not match. Defaults to (None, None).

        Returns:
            _type_: tuple
        """
        trie = self
        value_node = self.value_node
        value = self.not_set
        dep = 0
        for char in seq:
            if char in trie:
                dep += 1
                trie = trie[char]
                try:
                    value = trie[value_node]
                except KeyError:
                    pass
            else:
                break
        if value is not self.not_set:
            return (dep, value)
        return default


class GuessExt(object):
    r"""Determines whether the input bytes of a file prefix indicate a compressed file format.

    >>> cg = GuessExt()
    >>> cg.get_ext(b"PK\x05\x06zipfiledemo")
    '.zip'
    """

    MAGIC = {
        b"PK\x03\x04": ".zip",
        b"PK\x03\x06": ".zip",
        b"PK\x03\x08": ".zip",
        b"PK\x05\x04": ".zip",
        b"PK\x05\x06": ".zip",
        b"PK\x05\x08": ".zip",
        b"PK\x07\x04": ".zip",
        b"PK\x07\x06": ".zip",
        b"PK\x07\x08": ".zip",
        b'"\xb5/\xfd': ".zst",
        b"#\xb5/\xfd": ".zst",
        b"$\xb5/\xfd": ".zst",
        b"%\xb5/\xfd": ".zst",
        b"&\xb5/\xfd": ".zst",
        b"'\xb5/\xfd": ".zst",
        b"(\xb5/\xfd": ".zst",
        b"LZIP": ".lz",
        b"MSCF": ".cab",
        b"ISc(": ".cab",
        b"\x1f\x8b\x08": ".gz",
        b"BZh": ".bz2",
        b"\xfd7zXZ\x00": ".xz",
        b"7z\xbc\xaf'\x1c": ".7z",
        b"Rar!\x1a\x07\x00": ".rar",
        b"Rar!\x1a\x07\x01": ".rar",
    }

    def __init__(self):
        self.trie = Trie(self.MAGIC)

    def get_ext(self, magic_bytes: bytes, default=""):
        dep, value = self.trie.match(magic_bytes)
        if dep is None:
            return default
        else:
            return value


def xor_encode_decode(data, key):
    r"""
    Perform XOR encryption or decryption on the given data using a provided key.

    This function encrypts or decrypts the data by performing an XOR operation
    between each byte of the data and the corresponding byte of the key. The key
    is repeated if necessary to cover the entire length of the data.

    Parameters:
    - data: The byte data to be encrypted or decrypted.
    - key: The key used for encryption or decryption, also a sequence of bytes.

    Returns:
    - The resulting byte data after encryption or decryption.

    Example:

        >>> original_data = b'Hello, World!'
        >>> key = b'secret'
        >>> encrypted_data = xor_encode_decode(original_data, key)
        >>> encrypted_data
        b';\x00\x0f\x1e\nXS2\x0c\x00\t\x10R'
        >>> decrypted_data = xor_encode_decode(encrypted_data, key)
        >>> decrypted_data == original_data
        True
    """
    # Extend the key to ensure its length is at least as long as the data
    extended_key = key * (len(data) // len(key) + 1)
    # Perform XOR operation between each byte of the data and the extended key,
    # and return the new sequence of bytes
    return bytes([b ^ k for b, k in zip(data, extended_key)])


def is_running_win32(pid: int):
    with os.popen('tasklist /fo csv /fi "pid eq %s"' % int(pid)) as f:
        f.readline()
        text = f.readline()
        return bool(text)


def is_running_linux(pid: int):
    try:
        os.kill(int(pid), 0)
        return True
    except OSError:
        return False
    except SystemError:
        return True


def is_running(pid):
    """Check if the given process ID is running.

    Parameters:
    pid -- The process ID to check.

    Returns:
    True if the process is running; False otherwise.

    Examples:
    >>> is_running(os.getpid() * 10)  # Assume process is not running
    False
    >>> is_running(os.getpid())  # Check if the current process is running
    True
    >>> is_running("not_a_pid")  # Invalid PID input should be handled and return False
    False

    """
    try:
        pid = int(pid)
    except ValueError:
        return False
    if sys.platform == "win32":
        return is_running_win32(pid)
    else:
        return is_running_linux(pid)


def set_pid_file(
    path: Optional[Union[str, Path]] = None,
    raise_error=False,
    default_dir="",
    default_name="",
    default_level=2,
):
    """Sets a PID file to prevent multiple instances of a script or process from running concurrently.
    If no path is specified, it constructs a default based on the calling script's location and naming convention.

    The function checks for an existing PID file and verifies if the process is still running.

    If the process is running and `raise_error` is True, it raises a RuntimeError; otherwise, it returns False.
    Upon successful setup, it writes the current process ID (PID) to the file and schedules the file for deletion upon exit.

    Parameters:
    - path (Optional[Union[str, Path]]): The desired path for the PID file. Defaults to a generated path based on caller.
    - raise_error (bool): Determines whether to raise an error if the PID file is already locked. Defaults to False.
    - default_dir (str): The default directory for the PID file if none is provided. Defaults to "/dev/shm" or temp dir.
    - default_name (str): The default base name for the PID file. Automatically generated with the caller filename if not provided.
    - default_level (int): The number of directory levels to include in the default name from the caller's path.

    Returns:
    - Path: The path of the PID file if successfully created or updated.
    - False: If the PID file exists and the associated process is running, and `raise_error` is False.

    Raises:
    - ValueError: If unable to determine the caller's path or set a default name for the PID file.
    - RuntimeError: If `raise_error` is True and the PID file is locked by another process.

    Examples:

    >>> path = set_pid_file()  # Assuming this is the first run, should succeed
    >>> path.name.endswith('doctest.py.pid')
    True
    >>> set_pid_file()  # Simulating second instance trying to start, should raise error if raise_error=True
    False
    >>> path.unlink()
    """
    if path is None:
        if not default_name:
            try:
                stack = traceback.extract_stack()
                for index in range(len(stack) - 2, -1, -1):
                    caller_path = Path(stack[index].filename)
                    # print(caller_path)
                    if caller_path.is_file():
                        break
                else:
                    raise ValueError("Could not find caller path")
                if not caller_path.is_file():
                    pass
                parts: list = []
                for index, part in enumerate(caller_path.as_posix().split("/")[::-1]):
                    if index < default_level:
                        parts.insert(0, part)
                    else:
                        break
                default_name = "__".join(parts) + ".pid"
            except Exception as e:
                raise ValueError("Could not set default name for PID file: %s" % e)
        if not default_dir:
            default_dir = Path("/dev/shm")
            if not default_dir.is_dir():
                default_dir = Path(tempfile.gettempdir())
        path = Path(default_dir) / default_name
    if not isinstance(path, Path):
        path = Path(path)
    running = False
    if path.is_file():
        try:
            old_pid = int(path.read_text().strip())
            running = is_running(old_pid)
        except ValueError:
            # NaN pid
            pass

    if running:
        if raise_error:
            raise RuntimeError(f"{path.as_posix()} is locked by {old_pid}")
        else:
            return False
    else:
        pid_str = str(os.getpid())
        path.write_text(pid_str)
        new_pid = path.read_text().strip()
        if pid_str != new_pid:
            if raise_error:
                raise RuntimeError(
                    f"{path.as_posix()} is locked by other process {new_pid}"
                )
            else:
                return False

        def _release_file():
            if path.is_file() and path.read_text() == pid_str:
                path.unlink()

        atexit.register(_release_file)
        return path


def get_paste() -> Union[str, None]:
    """This module offers a simple utility for retrieving text from the system clipboard with tkinter.

    Function:
        get_paste() -> Union[str, None]

    Usage Note:
        While this function handles basic clipboard retrieval, for more advanced scenarios such as setting clipboard content or maintaining a persistent application interface, consider using libraries like `pyperclip` or running `Tkinter.mainloop` which keeps the GUI event loop active.
        Set clipboard with tkinter:
            _tk.clipboard_clear()
            _tk.clipboard_append(text)
            _tk.update()
            _tk.mainloop() # this is needed
    """
    from tkinter import TclError, Tk

    text = _tk = None
    try:
        _tk = Tk()
        _tk.withdraw()
        text = _tk.clipboard_get()
    except TclError:
        pass
    finally:
        if _tk:
            _tk.destroy()
        return text


def set_clip(text: str):
    """Copies the given text to the clipboard using a temporary file in a Windows environment.

    This function writes the provided text into a temporary file and then uses the `clip.exe` command-line utility
    to read from this file and copy its content into the clipboard.

    Parameters:
    text: str - The text content to be copied to the clipboard.
    """
    if sys.platform != "win32":
        raise RuntimeError("set_clip is only supported on Windows, you need pyperclip")
    try:
        path = Path(tempfile.gettempdir()) / "set_clip.txt"
        path.write_text(text)
        subprocess.run(
            f'clip.exe < "{path.name}"',
            cwd=path.parent.absolute(),
            shell=True,
        )
    finally:
        path.unlink(missing_ok=True)


class Clipboard:
    try:
        import pyperclip

        copy = pyperclip.copy
        paste = pyperclip.paste
    except ImportError:
        copy = set_clip
        paste = get_paste


def switch_flush_print():
    """Set builtins.print default flush=True.

    >>> print.__name__
    'print'
    >>> switch_flush_print()
    >>> print.__name__
    'flush_print'
    >>> switch_flush_print()
    >>> print.__name__
    'print'
    """
    import builtins

    if hasattr(builtins, "orignal_print"):
        orignal_print = builtins.orignal_print
        if builtins.print is not orignal_print:
            # back to orignal_print
            builtins.print = orignal_print
            return
    else:
        orignal_print = builtins.print
        builtins.orignal_print = orignal_print

    def flush_print(*args, **kwargs):
        if "flush" not in kwargs:
            kwargs["flush"] = True
        return orignal_print(*args, **kwargs)

    # set new flush_print
    builtins.print = flush_print


def unix_rlimit(max_mem: Optional[int] = None, max_file_size: Optional[int] = None):
    "Unix only. RLIMIT_RSS, RLIMIT_FSIZE to limit the max_memory and max_file_size"
    if sys.platform == "win32":
        raise RuntimeError("set_clip is only supported on Windows")
    import resource

    if max_mem is not None:
        resource.setrlimit(resource.RLIMIT_RSS, (max_mem, max_mem))
    if max_file_size is not None:
        resource.setrlimit(resource.RLIMIT_FSIZE, (max_file_size, max_file_size))


class SimpleFilter:
    """Simple dup-filter with pickle file.

    >>> for r in range(1, 5):
    ...     try:
    ...         done = 0
    ...         with SimpleFilter("1.proc", 0, success_unlink=True) as sf:
    ...             for i in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
    ...                 if sf.exist(i):
    ...                     continue
    ...                 if done > 2:
    ...                     print("crash!")
    ...                     raise ValueError
    ...                 print('[round %s]' % r, i, 'done', flush=True)
    ...                 sf.add(i)
    ...                 done += 1
    ...             break
    ...     except ValueError:
    ...         continue
    ...
    [round 1] 1 done
    [round 1] 2 done
    [round 1] 3 done
    crash!
    [round 2] 4 done
    [round 2] 5 done
    [round 2] 6 done
    crash!
    [round 3] 7 done
    [round 3] 8 done
    [round 3] 9 done
    crash!
    [round 4] 10 done
    """

    def __init__(self, path: Union[Path, str], delay=2, success_unlink=False):
        self.path = Path(path)
        self.delay = delay
        self.success_unlink = success_unlink

        self.timer = None
        self.has_new = False
        self.w_lock = Lock()
        self.current_thread = current_thread()

    def add(self, key):
        self.has_new = True
        self.cache.add(key)
        if not self.timer:
            self.timer = Timer(self.delay, self.save_cache)
            self.timer.start()

    def exist(self, key):
        return key in self.cache

    def save_cache(self):
        with self.w_lock:
            if self.has_new:
                self.path.write_bytes(pickle.dumps(self.cache))
            if self.timer:
                if current_thread() is not self.current_thread:
                    # timer finished
                    self.timer = None

    def __enter__(self):
        try:
            self.cache = pickle.loads(self.path.read_bytes())
        except Exception:
            self.cache = set()
            self.save_cache()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.timer and not self.timer.is_alive():
            self.timer.cancel()
        if exc_type is None:
            # success & finish
            if self.success_unlink:
                self.has_new = False
                with self.w_lock:
                    # remove cache file
                    self.path.unlink(missing_ok=True)
                return
        self.save_cache()


class FileDict(dict):
    """A dict that can be saved to a file.

    Demo::

        >>> d = FileDict.load("test.json")
        >>> d.get('a', 'none')
        'none'
        >>> d['a'] = 1
        >>> d.save()
        >>> d2 = FileDict.load("test.json")
        >>> d2.get('a')
        1
        >>> d2.unlink()
        >>> d = FileDict.load("test.pkl")
        >>> d.get('a', 'none')
        'none'
        >>> d['a'] = 1
        >>> d.save()
        >>> d2 = FileDict.load("test.pkl")
        >>> d2.get('a')
        1
        >>> d2.unlink()
    """

    @classmethod
    def choose_handler(cls, filename):
        ext = os.path.splitext(filename)[1]
        if ext == ".json":
            return cls.json_handler
        elif ext == ".pkl":
            return cls.pickle_handler
        raise ValueError("unknown file extension")

    @staticmethod
    def json_handler(path: str, obj=None):
        if isinstance(obj, dict):
            temp = path + ".cache"
            try:
                with open(temp, "w") as f:
                    json.dump(obj, f)
                try:
                    os.unlink(path)
                except FileNotFoundError:
                    pass
                os.rename(temp, path)
            finally:
                try:
                    os.unlink(temp)
                except FileNotFoundError:
                    pass
        else:
            try:
                with open(path, "r") as f:
                    return json.load(f)
            except FileNotFoundError:
                return {}

    @staticmethod
    def pickle_handler(path: str, obj=None):
        if isinstance(obj, dict):
            temp = path + ".cache"
            try:
                with open(temp, "wb") as f:
                    pickle.dump(obj, f)
                try:
                    os.unlink(path)
                except FileNotFoundError:
                    pass
                os.rename(temp, path)
            finally:
                try:
                    os.unlink(temp)
                except FileNotFoundError:
                    pass
        else:
            try:
                with open(path, "rb") as f:
                    return pickle.load(f)
            except FileNotFoundError:
                return {}

    @classmethod
    def load(cls, path: str, handler=None):
        path = str(path)
        if not handler:
            handler = cls.choose_handler(str(path))
        self = cls(handler(path))
        setattr(self, "path", path)
        setattr(self, "handler", handler)
        return self

    def save(self):
        self.handler(self.path, self)

    def unlink(self):
        try:
            os.unlink(self.path)
        except FileNotFoundError:
            pass


def i2b(n: int, length=0, byteorder="big", signed=False) -> bytes:
    r"""Convert an int to bytes of a specified length, commonly used in TCP communication.

    Parameters:
        n: The integer to convert.
        length: The number of bytes to convert. Default is 0, which means the byte length is determined automatically based on the integer's bit length.
        byteorder: The byte order, which can be "big" or "little". Default is "big".
        signed: Whether the integer is signed. Default is False.

    Returns:
        The converted byte sequence.

    Length  Maximum::

        1   256B-1
        2    64K-1
        3    16M-1
        4     4G-1
        5     1T-1
        6    64T-1
        7    16E-1
        8   256Z-1
        9    32Y-1
        10    4P-1

    >>> i2b(0)
    b''
    >>> i2b(1)
    b'\x01'
    >>> i2b(1, length=2)
    b'\x00\x01'
    >>> i2b(255)
    b'\xff'
    >>> i2b(256)
    b'\x01\x00'
    >>> i2b(256, length=3)
    b'\x00\x01\x00'
    >>> i2b(256, byteorder="little")
    b'\x00\x01'
    >>> i2b(256, length=3, signed=True)
    b'\x00\x01\x00'
    """
    if not length:
        if signed:
            raise ValueError("signed must be False when length is 0")
        length = (n.bit_length() + 7) // 8
    return n.to_bytes(length, byteorder=byteorder, signed=signed)


def b2i(b: bytes, byteorder="big", signed=False) -> int:
    r"""Convert a byte sequence to an integer.

    Parameters:
        b: The byte sequence to convert.
        byteorder: The byte order, which can be "big" or "little". Default is "big".
        signed: Whether the integer is signed. Default is False.

    Returns:
        The converted integer.
    >>> b2i(b'\x01')
    1
    >>> b2i(b'\x00\x01')
    1
    >>> b2i(b'\x00\x01', byteorder="little")
    256
    >>> b2i(b'\xff')
    255
    >>> b2i(b'\x01\x00')
    256
    >>> b2i(b'\x00\x01\x00')
    256
    >>> b2i(b'\x00\x01\x00', signed=True)
    256
    >>> b2i(b'\x00\x01\x00', signed=False)
    256
    """
    return int.from_bytes(b, byteorder=byteorder, signed=signed)


class PathLock:
    """A Lock/asyncio.Lock of a path, and the child-path lock will block the parent-path.
    Ensure a path and its child-path are not busy. Can be used in a with statement to avoid race condition, such as rmtree.

    Demo::

        import asyncio
        import time
        from threading import Thread

        from morebuiltins.utils import Path, PathLock


        def test_sync_lock():
            parent = Path("/tmp")
            child = Path("/tmp/child")
            result = []

            def parent_job():
                with PathLock(parent):
                    print(time.strftime("%Y-%m-%d %H:%M:%S"), "parent done", flush=True)
                    result.append(parent)

            def child_job():
                with PathLock(child):
                    time.sleep(0.5)
                    print(time.strftime("%Y-%m-%d %H:%M:%S"), "child done", flush=True)
                    result.append(child)

            Thread(target=child_job).start()
            time.sleep(0.01)
            t = Thread(target=parent_job)
            t.start()
            t.join()
            # 2024-07-16 22:52:20 child done
            # 2024-07-16 22:52:20 parent done
            # child before parent
            assert result == [child, parent], result


        async def test_async_lock():
            parent = Path("/tmp")
            child = Path("/tmp/child")
            result = []

            async def parent_job():
                async with PathLock(parent):
                    print(time.strftime("%Y-%m-%d %H:%M:%S"), "parent done", flush=True)
                    result.append(parent)

            async def child_job():
                async with PathLock(child):
                    await asyncio.sleep(0.5)
                    print(time.strftime("%Y-%m-%d %H:%M:%S"), "child done", flush=True)
                    result.append(child)

            asyncio.create_task(child_job())
            await asyncio.sleep(0.01)
            await asyncio.create_task(parent_job())
            # 2024-07-16 22:52:21 child done
            # 2024-07-16 22:52:21 parent done
            # child before parent
            assert result == [child, parent], result


        test_sync_lock()
        asyncio.run(test_async_lock())
    """

    GLOBAL_LOCK: Optional[Lock] = None
    LOCKS: List[Tuple[Path, Lock]] = []
    GLOBAL_ALOCK: Optional[asyncio.Lock] = None
    ALOCKS: List[Tuple[Path, asyncio.Lock]] = []

    def __init__(self, path: Path, timeout=None):
        # ensure resolved, avoid the ../../ or softlink problem
        self.path = Path(path).resolve()
        self.timeout = timeout
        if self.GLOBAL_LOCK is None:
            self.__class__.GLOBAL_LOCK = Lock()

    @property
    def global_lock(self):
        if self.__class__.GLOBAL_LOCK is None:
            self.__class__.GLOBAL_LOCK = Lock()
        return self.__class__.GLOBAL_LOCK

    @property
    def global_alock(self):
        if self.__class__.GLOBAL_ALOCK is None:
            self.__class__.GLOBAL_ALOCK = asyncio.Lock()
        return self.__class__.GLOBAL_ALOCK

    def get_lock(self) -> Lock:
        with self.global_lock:
            for path, lock in self.LOCKS:
                if path.is_relative_to(self.path):
                    self.LOCKS.append((self.path, lock))
                    return lock
            lock = Lock()
            self.LOCKS.append((self.path, lock))
            return lock

    async def get_alock(self) -> asyncio.Lock:
        async with self.global_alock:
            for path, lock in self.ALOCKS:
                if path.is_relative_to(self.path):
                    self.ALOCKS.append((self.path, lock))
                    return lock
            lock = asyncio.Lock()
            self.ALOCKS.append((self.path, lock))
            return lock

    def __enter__(self):
        self.lock = self.get_lock()
        if self.timeout:
            self.lock.acquire(blocking=True, timeout=self.timeout)
        else:
            self.lock.acquire(blocking=True)

    def __exit__(self, *_):
        with self.GLOBAL_LOCK:
            try:
                self.LOCKS.remove(self.path)
            except ValueError:
                pass
        self.lock.release()

    async def __aenter__(self):
        self.lock = await self.get_alock()
        await self.lock.acquire()

    async def __aexit__(self, *_):
        with self.GLOBAL_LOCK:
            try:
                self.ALOCKS.remove(self.path)
            except ValueError:
                pass
        self.lock.release()


def get_size(obj, seen=None, iterate_unsafe=False) -> int:
    """Recursively get size of objects.

    Args:
        obj: object of any type
        seen (set): set of ids of objects already seen
        iterate_unsafe (bool, optional): whether to iterate through generators/iterators. Defaults to False.

    Returns:
        int: size of object in bytes

    Examples:
    >>> get_size("") > 0
    True
    >>> get_size([]) > 0
    True
    >>> def gen():
    ...     for i in range(10):
    ...         yield i
    >>> g = gen()
    >>> get_size(g) > 0
    True
    >>> next(g)
    0
    >>> get_size(g, iterate_unsafe=True) > 0
    True
    >>> try:
    ...     next(g)
    ... except StopIteration:
    ...     "StopIteration"
    'StopIteration'
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    seen.add(obj_id)
    if isinstance(obj, (str, bytes, bytearray)):
        pass
    elif isinstance(obj, dict):
        size += sum([get_size(v, seen, iterate_unsafe) for v in obj.values()])
        size += sum([get_size(k, seen, iterate_unsafe) for k in obj.keys()])
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen, iterate_unsafe)
    elif isinstance(obj, types.GeneratorType) or isinstance(
        obj, collections.abc.Iterator
    ):
        if iterate_unsafe:
            # Warning: this will consume the generator/iterator
            size += sum([get_size(i, seen, iterate_unsafe) for i in obj])
    elif hasattr(obj, "__iter__"):
        # Safe to iterate through containers like lists, tuples, sets, etc.
        size += sum([get_size(i, seen, iterate_unsafe) for i in obj])
    return size


def base_encode(
    num: int,
    alphabet: str = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz",
) -> str:
    """Encode a number to a base-N string.

    Args:
        num (int): The number to encode.
        alphabet (str, optional): Defaults to "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz", which follows the order of ASCII characters.

    Returns:
        str: The encoded string.

    Examples:
    >>> base_encode(0)
    '0'
    >>> base_encode(1)
    '1'
    >>> base_encode(10000000000000)
    '2q3Rktoe'
    >>> base_encode(10000000000000, "0123456789")
    '10000000000000'
    >>> a = [base_encode(i).zfill(10) for i in range(10000)]
    >>> sorted(a) == a
    True
    """
    length = len(alphabet)
    result = ""
    while num:
        num, i = divmod(num, length)
        result = f"{alphabet[i]}{result}"
    return result or alphabet[0]


def base_decode(
    string: str,
    alphabet: str = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
) -> int:
    """Decode a base-N string to a number.

    Args:
        string (str): The string to decode.
        alphabet (str, optional): Defaults to "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".

    Returns:
        int: The decoded number.

    Examples:
    >>> base_decode("0")
    0
    >>> base_decode("1")
    1
    >>> base_decode("2Q3rKTOE")
    10000000000000
    >>> base_decode("10000000000000", "0123456789")
    10000000000000
    """
    length = len(alphabet)
    result = 0
    for char in string:
        result = result * length + alphabet.index(char)
    return result


def gen_id(rand_len=4) -> str:
    """Generate a unique ID based on the current time and random bytes

    Args:
        rand_len (int, optional): Defaults to 4.

    Returns:
        str: The generated ID.

    Examples:
    >>> a, b = gen_id(), gen_id()
    >>> a != b
    True
    >>> import time
    >>> ids = [time.sleep(0.000001) or gen_id() for _ in range(1000)]
    >>> len(set(ids))
    1000
    """
    now = datetime.now()
    s1 = now.strftime("%y%m%d_%H%M%S")
    s2 = base_encode(now.microsecond)
    s2 = f"{s2:>04}{os.urandom(rand_len // 2).hex()}"
    return f"{s1}_{s2}"


def timeti(
    stmt: Union[str, Callable] = "pass",
    setup="pass",
    timer=timeit.default_timer,
    number=1000000,
    globals=None,
) -> int:
    """Return the number of iterations per second for a given statement.

    Args:
        stmt (str, optional): Defaults to "pass".
        setup (str, optional): Defaults to "pass".
        timer (optional): Defaults to timeit.default_timer.
        number (int, optional): Defaults to 1000000.
        globals (dict, optional): Defaults to None.

    Returns:
        int: The number of iterations per second.

    Examples:
    >>> timeti("1 / 1") > 1000000
    True
    >>> timeti(lambda : 1 + 1, number=100000) > 100000
    True
    """
    result = timeit.timeit(
        stmt=stmt, setup=setup, timer=timer, number=number, globals=globals
    )
    return int(1 / (result / number))


class SnowFlake:
    def __init__(
        self,
        machine_id: int = 1,
        worker_id: int = 1,
        wait_for_next_ms=True,
        start_epoch_ms: int = -1,
        *,
        max_bits=64,
        sign_bit=1,
        machine_id_bit=5,  # between 0 and 31
        worker_id_bit=5,  # between 0 and 31
        seq_bit=12,  # between 0 and 4095
        start_date="2025-01-01",
    ):
        r"""Generate unique IDs using Twitter's Snowflake algorithm.

        The ID is composed of:
        - 41 bits timestamp in milliseconds since a custom epoch
        - 5 bits machine ID
        - 5 bits worker ID
        - 12 bits sequence number
        If you need to ensure thread safety, please lock it yourself.

            Args:
                machine_id (int): the machine_id of the SnowFlake object
                worker_id (int): the worker_id of the SnowFlake object
                wait_for_next_ms (bool, optional): whether to wait for next millisecond if sequence overflows. Defaults to True.
                start_epoch_ms (int, optional): the start epoch in milliseconds. Defaults to -1.
                *
                max_bits (int, optional): the maximum bits of the ID. Defaults to 64.
                sign_bit (int, optional): the sign bit of the ID. Defaults to 1.
                machine_id_bit (int, optional): the machine_id bit of the ID. Defaults to 5.
                worker_id_bit (int, optional): the worker_id bit of the ID. Defaults to 5.
                seq_bit (int, optional): the sequence bit of the ID. Defaults to 12.
                start_date (str): the start date of the SnowFlake object. Defaults to "2025-01-01", only used when start_epoch_ms is -1.

            Example:
            >>> import time
            >>> snowflake = SnowFlake()
            >>> ids = [snowflake.get_id() for _ in range(10000)]
            >>> len(set(ids)) == len(ids)
            True
            >>> # test timestamp overflow
            >>> snowflake = SnowFlake(1, 1, start_date=time.strftime("%Y-%m-%d"))
            >>> timeleft = snowflake.timestamp_overflow_check() // 1000 // 60 // 60 // 24 // 365
            >>> timeleft == 69
            True
            >>> snowflake = SnowFlake(1, 1, start_date=time.strftime("%Y-%m-%d"), sign_bit=0)
            >>> timeleft = snowflake.timestamp_overflow_check() // 1000 // 60 // 60 // 24 // 365
            >>> timeleft >= 138
            True
            >>> # test machine_id and worker_id overflow
            >>> try:
            ...     snowflake = SnowFlake(32, 32)
            ... except ValueError as e:
            ...     e
            ValueError('Machine ID must be between 0 and 31')
            >>> sf = SnowFlake(32, 32, machine_id_bit=6, worker_id_bit=6)
            >>> sf.max_machine_id
            63
            >>> sf = SnowFlake(32, machine_id_bit=64)
            >>> sf.timestamp_overflow_check() < 0
            True
        """
        self.max_bits = max_bits
        self.sign_bit = sign_bit
        self.machine_id_bit = machine_id_bit
        self.worker_id_bit = worker_id_bit
        self.seq_bit = seq_bit
        self.max_seq = (1 << self.seq_bit) - 1
        self.max_worker_id = (1 << self.worker_id_bit) - 1
        self.max_machine_id = (1 << self.machine_id_bit) - 1
        self.max_timestamp = self.get_max_timestamp()
        # Validate inputs
        if not 0 <= machine_id <= self.max_machine_id:
            raise ValueError(f"Machine ID must be between 0 and {self.max_machine_id}")
        if not 0 <= worker_id <= self.max_worker_id:
            raise ValueError(f"Worker ID must be between 0 and {self.max_worker_id}")
        self.timestamp_shift = self.seq_bit + self.worker_id_bit + self.machine_id_bit
        self.machine_id = machine_id
        self.worker_id = worker_id
        # Calculate parts of ID
        self.machine_id_part = machine_id << (self.seq_bit + self.worker_id_bit)
        self.worker_id_part = worker_id << self.seq_bit
        # Initialize sequence, timestamp and last timestamp
        self.seq = 0
        self.last_timestamp = -1
        if start_epoch_ms < 0:
            self.start_epoch_ms = self.str_to_ms(start_date)
        else:
            self.start_epoch_ms = start_epoch_ms
        self.wait_for_next_ms = wait_for_next_ms

    def get_max_timestamp(self):
        """Get maximum timestamp that can be represented"""
        return (1 << (self.max_bits - self.sign_bit)) - 1 >> (
            self.seq_bit + self.worker_id_bit + self.machine_id_bit
        )

    def timestamp_overflow_check(self):
        """Check how many milliseconds left until timestamp overflows"""
        return self.get_max_timestamp() - self._ms_passed()

    @staticmethod
    def str_to_ms(string: str) -> int:
        """Convert string to milliseconds since start_time"""
        return int(mktime(strptime(string, "%Y-%m-%d")) * 1000)

    def _ms_passed(self):
        """Get current timestamp in milliseconds since start_time"""
        return int(time() * 1000 - self.start_epoch_ms)

    def _wait_next_millis(self, last_timestamp):
        """Wait until next millisecond"""
        if not self.wait_for_next_ms:
            raise RuntimeError(
                f"Over {self.max_seq} IDs generated in 1ms, increase the wait_for_next_ms parameter"
            )
        timestamp = self._ms_passed()
        while timestamp <= last_timestamp:
            timestamp = self._ms_passed()
        return timestamp

    def get_id(self):
        """Generate next unique ID"""
        now = self._ms_passed()

        # Clock moved backwards, reject requests
        if now < self.last_timestamp:
            raise RuntimeError(
                f"Clock moved backwards. Refusing to generate ID for {self.last_timestamp - now} milliseconds"
            )

        # Same timestamp, increment sequence
        if now == self.last_timestamp:
            self.seq = (self.seq + 1) & self.max_seq
            # Sequence overflow, wait for next millisecond
            if self.seq == 0:
                now = self._wait_next_millis(now)
        else:
            # Reset sequence for different timestamp
            self.seq = 0

        self.last_timestamp = now
        # Compose ID from components
        return (
            (now << self.timestamp_shift)
            | (self.machine_id_part)
            | (self.worker_id_part)
            | self.seq
        )


def cut_file(path_or_file, max_bytes, remain_ratio=0.5, ensure_line_start=False):
    r"""Cut file to max_bytes, remain_ratio is the ratio of the end part to remain, ensure_line_start is to ensure the first line is a complete line

    Args:
       path_or_file (str/Path/file): input file path or file-like object
       max_bytes (int): max bytes before cut
       remain_ratio (float, optional): remain ratio of the end part. Defaults to 0.5.
       ensure_line_start (bool, optional): ensure the first line is a complete line. Defaults to False.

    Examples:
    >>> from io import BytesIO
    >>> f = BytesIO(b"1234567890\n1234567890\n1234567890\n")
    >>> cut_file(f, 30)
    >>> f.getvalue()
    b'890\n1234567890\n'
    >>> f = BytesIO(b"1234567890\n1234567890\n1234567890\n")
    >>> cut_file(f, 30, ensure_line_start=True)
    >>> f.getvalue()
    b'1234567890\n'
    >>> f = BytesIO(b"1234567890\n1234567890\n1234567890\n")
    >>> cut_file(f, 30, remain_ratio=0)
    >>> f.getvalue()
    b''
    """
    need_close = False
    try:
        if isinstance(path_or_file, (str, Path)):
            f = open(path_or_file, "r+b")
            need_close = True
        elif hasattr(path_or_file, "read"):
            f = path_or_file
        else:
            raise TypeError("path_or_file must be str, Path or file-like object")
        size = f.seek(0, 2)
        if size > max_bytes:
            f.seek(int(size - max_bytes * remain_ratio))
            left = f.read()
            if ensure_line_start:
                left = re.sub(b"^[^\n]+\n", b"", left, count=1)
            f.seek(0)
            f.truncate()
            f.write(left)
            f.flush()
    finally:
        if need_close:
            f.close()


if __name__ == "__main__":
    __name__ = "morebuiltins.utils"
    import doctest

    doctest.testmod()



import ast
import re
import sys
from pathlib import Path

__all__ = ["parse_deps"]

try:
    stdlibs = set(sys.stdlib_module_names)
except AttributeError:
    # fix python3.9 sys.stdlib_module_names not found
    import os
    import pkgutil

    stdlib_path = os.path.dirname(os.__file__)
    stdlibs = {name for _, name, _ in pkgutil.iter_modules([stdlib_path])}

stds_libs = stdlibs | set(sys.builtin_module_names)


def parse_deps(project_dir, ignore_stds=True, format_path=True, pattern_list=("*.py",)):
    r"""Parse dependencies of a project directory, and find circular dependencies.

    Args:
          project_dir (str): Path to the project directory.
          ignore_stds (bool, optional): Whether to ignore dependencies from the standard library. Defaults to True.
          format_path (bool, optional): Whether to format the paths. Defaults to True.
          pattern_list (tuple, optional): List of patterns to match files. Defaults to ("*.py",).
    Returns:
            dict: A dictionary containing the project directory, circular dependencies, and dependencies.

    Demo::

        import json
        import multiprocessing
        from pathlib import Path

        project_dir = Path(multiprocessing.__file__).parent
        result = parse_deps(
            project_dir,
            ignore_stds=True,
            format_path=True,
            pattern_list=("*.py",),
        )
        dependencies = sorted(
            result["dependencies"].items(),
            key=lambda i: (len(i[1]), i[0]),
            reverse=True,
        )
        print("project_dir:", project_dir.as_posix(), flush=True)
        print("circular_dependency:", result["circular_dependency"], flush=True)
        for source, deps in dependencies:
            print(source, f"({len(deps)})", flush=True)
            for i in deps:
                print("\t", i, flush=True)
        # project_dir: D:/python311/Lib/multiprocessing
        # circular_dependency: [('./connection.py', './context.py'), ('./context.py', './forkserver.py'), ('./context.py', './managers.py'), ('./context.py', './popen_forkserver.py'), ('./context.py', './popen_spawn_posix.py'), ('./context.py', './popen_spawn_win32.py'), ('./context.py', './sharedctypes.py'), ('./context.py', './spawn.py'), ('./dummy/__init__.py', './pool.py')]
        # ./context.py (13)
        # 	 ./connection.py
        # 	 ./forkserver.py
        # 	 ./managers.py
        # 	 ./pool.py
        # 	 ./popen_fork.py
        # 	 ./popen_forkserver.py
        # 	 ./popen_spawn_posix.py
        # 	 ./popen_spawn_win32.py
        # 	 ./queues.py
        # 	 ./sharedctypes.py
        # 	 ./spawn.py
        # 	 ./synchronize.py
        # 	 ./util.py
        # ./synchronize.py (2)
        # 	 ./heap.py
        # 	 ./resource_tracker.py
        # ./resource_sharer.py (2)
        # 	 ./connection.py
        # 	 ./context.py
        # ./queues.py (2)
        # 	 ./synchronize.py
        # 	 ./util.py
        # ./pool.py (2)
        # 	 ./connection.py
        # 	 ./dummy/__init__.py
        # ./dummy/__init__.py (2)
        # 	 ./dummy/connection.py
        # 	 ./pool.py
        # ./util.py (1)
        # 	 test
        # ./spawn.py (1)
        # 	 ./context.py
        # ./sharedctypes.py (1)
        # 	 ./context.py
        # ./reduction.py (1)
        # 	 ./resource_sharer.py
        # ./process.py (1)
        # 	 ./context.py
        # ./popen_spawn_win32.py (1)
        # 	 ./context.py
        # ./popen_spawn_posix.py (1)
        # 	 ./context.py
        # ./popen_forkserver.py (1)
        # 	 ./context.py
        # ./managers.py (1)
        # 	 ./context.py
        # ./heap.py (1)
        # 	 ./context.py
        # ./forkserver.py (1)
        # 	 ./context.py
        # ./connection.py (1)
        # 	 ./context.py

    """

    def get_dependencies(file_path: Path, ignore_stds=True):
        dependencies: set = set()
        try:
            tree = ast.parse(file_path.read_text("utf-8", "ignore"))
        except SyntaxError:
            return dependencies
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    name = alias.name
                    if name and name.split(".", 1)[0] not in stds_libs:
                        dependencies.add(name)
            elif isinstance(node, ast.ImportFrom):
                name = node.module or ""
                if not name:
                    continue
                if ignore_stds and name.split(".", 1)[0] in stds_libs:
                    continue
                dependencies.add("." * (node.level) + name)
        return dependencies

    def fix_path(file_path: Path, name: str, project_dir: Path):
        match = re.match(r"^(\.+)(.+)", name)
        if match:
            # a/b/c.py + ..d => a/d.py
            # fix relative import .. to absolute import
            level = len(match[1])
            tail = match[2]
            base = file_path
            for _ in range(level):
                base = base.parent
            name = ".".join([base.name, tail])
        # name to path
        path_tail = name.replace(".", "/") + ".py"
        head_name, _, tail = path_tail.partition("/")
        dir_path = file_path.parent
        for _ in range(len(file_path.parts)):
            if dir_path.name == head_name:
                result_path = dir_path.joinpath(tail)
                if result_path.is_file():
                    result = "./" + result_path.relative_to(project_dir).as_posix()
                    return result
                else:
                    maybe_dir = result_path.with_suffix("")
                    if maybe_dir.is_dir():
                        maybe_file = maybe_dir.joinpath("__init__.py")
                        if maybe_file.is_file():
                            result = (
                                "./" + maybe_file.relative_to(project_dir).as_posix()
                            )
                            return result
            dir_path = dir_path.parent
        return name

    project_dir = Path(project_dir)
    if not project_dir.is_dir():
        raise ValueError(f"{project_dir} is not a directory")
    dependencies = {}
    for pattern in pattern_list:
        for file_path in project_dir.rglob(pattern):
            deps = get_dependencies(file_path, ignore_stds=ignore_stds)
            if deps:
                key = "./" + file_path.relative_to(project_dir).as_posix()
                if key in dependencies:
                    continue
                if format_path:
                    deps = {fix_path(file_path, name, project_dir) for name in deps}
                    deps = {i for i in deps if i}
                    if not deps:
                        continue
                dependencies[key] = sorted(deps, key=lambda i: (i.startswith("./"), i))
    cur_deps = set()
    for source, deps in dependencies.items():
        for dep in deps:
            if source in dependencies.get(dep, []):
                cur_deps.add(tuple(sorted([source, dep])))
    cur_deps = sorted(cur_deps)
    result = {
        "project_dir": project_dir.resolve().as_posix(),
        "circular_dependency": cur_deps,
        "dependencies": dependencies,
    }
    return result


def main():
    import argparse
    import json

    parser = argparse.ArgumentParser()

    parser.add_argument("--project-dir", default=r".")
    parser.add_argument(
        "--ignore-stds",
        type=int,
        default=1,
        choices=[0, 1],
        help="Whether to ignore dependencies from the standard library, default to 1",
    )
    parser.add_argument(
        "--format-path",
        type=int,
        default=1,
        choices=[0, 1],
        help="Whether to format the paths, default to 1",
    )
    parser.add_argument(
        "--pattern-list",
        nargs="+",
        default=("*.py",),
        help="List of patterns to match files, default to ['*.py']",
    )
    parser.add_argument("--print-fmt", choices=["default", "json"], default="default")
    args = parser.parse_args()
    project_dir = Path(args.project_dir)
    result = parse_deps(
        project_dir,
        ignore_stds=args.ignore_stds,
        format_path=args.format_path,
        pattern_list=args.pattern_list,
    )

    if args.print_fmt == "json":
        print(json.dumps(result, ensure_ascii=False), flush=True)
    else:
        dependencies = sorted(
            result["dependencies"].items(),
            key=lambda i: (len(i[1]), i[0]),
            reverse=True,
        )
        print("project_dir:", project_dir.as_posix(), flush=True)
        print("circular_dependency:", result["circular_dependency"], flush=True)
        for source, deps in dependencies:
            print(source, f"({len(deps)})", flush=True)
            for i in deps:
                print("\t", i, flush=True)


if __name__ == "__main__":
    main()



import argparse
import subprocess
import sys
from pathlib import Path
from typing import Any, Dict

__all__ = ["service_handler"]


def run_systemctl(command: str, service_name: str) -> int:
    """Execute systemctl command"""
    cmd = ["systemctl", command, service_name]
    try:
        subprocess.run(cmd, check=True, capture_output=True, text=True)
        print(f"Successfully executed: systemctl {command} {service_name}")
        return 0
    except subprocess.CalledProcessError as e:
        print(f"Error: {e.stderr}")
        return 1


def manage_service_file(
    service_name: str, content: str = "", action: str = "create"
) -> int:
    """Manage systemd service file"""
    service_path = Path("/etc/systemd/system") / f"{service_name}.service"

    if action == "create":
        try:
            service_path.write_text(content)
            subprocess.run(["systemctl", "daemon-reload"])
            return 0
        except Exception as e:
            print(f"Error creating service file: {e}")
            return 1
    elif action == "delete":
        try:
            if service_path.exists():
                service_path.unlink()
                subprocess.run(["systemctl", "daemon-reload"])
            return 0
        except Exception as e:
            print(f"Error removing service file: {e}")
            return 1
    return 0


def create_service_file(args: Dict[str, Any]) -> str:
    """Generate systemd service file content"""
    # Validate required parameter
    if not args.get("ExecStart"):
        raise ValueError("Error: ExecStart is required for service file creation")

    content = ["[Unit]"]

    # Unit section configuration - keep only common params
    unit_params = ["Description", "After", "Requires", "Wants"]
    for param in unit_params:
        if args.get(param):
            content.append(f"{param}={args[param]}")

    content.append("\n[Service]")
    # Service section configuration - keep only common params
    service_params = [
        "Type",
        "ExecStart",
        "ExecStop",
        "User",
        "Group",
        "WorkingDirectory",
        "Restart",
        "RestartSec",
        "Environment",
        "EnvironmentFile",
        "StandardOutput",
        "StandardError",
    ]
    for param in service_params:
        value = args.get(param)
        if value:
            if isinstance(value, list):
                for v in value:
                    content.append(f"Environment={v}")
            else:
                content.append(f"{param}={args[param]}")

    content.append("\n[Install]")
    content.append(f"WantedBy={args.get('WantedBy', 'multi-user.target')}")

    # Handle custom parameters
    custom_sections: dict = {}
    for key, value in args.items():
        if "__" in key and value is not None:
            section, param = key.split("__", 1)
            if section not in custom_sections:
                custom_sections[section] = []
            custom_sections[section].append(f"{param}={value}")

    for section, params in custom_sections.items():
        if params:
            content.append(f"\n[{section}]")
            content.extend(params)

    return "\n".join(content)


def service_handler():
    """Generate and manage systemd service files

    Example usage:

    1. Create, enable and start service:
        python -m morebuiltins.cmd.systemd.service -name myservice -enable -Description "My service" -ExecStart "/bin/bash myscript.sh"
    2. Stop, disable and remove service:
        python -m morebuiltins.cmd.systemd.service -name myservice -disable
    3. Print service file content:
        python -m morebuiltins.cmd.systemd.service -name myservice -Description "My service" -ExecStart "/bin/bash myscript.sh" -Type simple
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter, usage=service_handler.__doc__
    )

    # Basic arguments
    parser.add_argument(
        "-name", "--name", help="Service name (required)", required=True
    )
    parser.add_argument(
        "-enable",
        "--enable",
        "-start",
        "--start",
        action="store_true",
        help="Create, enable and start service",
    )
    parser.add_argument(
        "-disable",
        "--disable",
        "-stop",
        "--stop",
        action="store_true",
        help="Stop, disable and remove service",
    )

    # Unit section - common arguments
    parser.add_argument("-Description", "--Description", help="Service description")
    parser.add_argument("-After", "--After", help="Start after specified units")
    parser.add_argument("-Requires", "--Requires", help="Required units")
    parser.add_argument("-Wants", "--Wants", help="Wanted units")

    # Service section - common arguments
    parser.add_argument(
        "-Type", "--Type", choices=["simple", "forking", "oneshot", "notify"]
    )
    parser.add_argument("-ExecStart", "--ExecStart", help="Start command")
    parser.add_argument("-ExecStop", "--ExecStop", help="Stop command")
    parser.add_argument("-User", "--User", help="User to run service")
    parser.add_argument("-Group", "--Group", help="Group to run service")
    parser.add_argument(
        "-WorkingDirectory", "--WorkingDirectory", help="Working directory"
    )
    parser.add_argument(
        "-Restart", "--Restart", choices=["no", "always", "on-success", "on-failure"]
    )
    parser.add_argument(
        "-RestartSec", "--RestartSec", help="Restart interval (seconds)"
    )
    parser.add_argument(
        "-Environment", "--Environment", help="Environment variables", action="append"
    )
    parser.add_argument(
        "-EnvironmentFile", "--EnvironmentFile", help="Environment file"
    )
    parser.add_argument(
        "-StandardOutput",
        "--StandardOutput",
        help="Standard output, e.g. syslog, journal, append:/tmp/app.log, file:/tmp/app.log",
        default="",
    )
    parser.add_argument(
        "-StandardError",
        "--StandardError",
        help="Standard error, e.g. syslog, journal, append:/tmp/app.log, file:/tmp/app.log",
        default="",
    )

    # Install section arguments
    parser.add_argument("-WantedBy", "--WantedBy", default="multi-user.target")
    parser.add_argument("-RequiredBy", "--RequiredBy", help="Required by units")
    parser.add_argument("-Also", "--Also", help="Also enable these units")
    parser.add_argument("-Alias", "--Alias", help="Unit aliases")

    # Support custom section parameters in format: -Section__Key=value
    # Example: -Network__IPForward=yes will add IPForward=yes under [Network] section
    args, unknown = parser.parse_known_args()

    # Handle unknown parameters
    for arg in unknown:
        if arg.startswith(("-", "--")) and "=" in arg:
            key = arg.lstrip("-").split("=")[0]
            value = arg.split("=")[1]
            setattr(args, key, value)

    if args.disable:
        run_systemctl("stop", f"{args.name}.service")
        run_systemctl("disable", f"{args.name}.service")
        return manage_service_file(args.name, action="delete")

    elif args.enable:
        if not args.ExecStart:
            print("Error: -ExecStart is required for enabling service")
            parser.print_help()
            return 1

        service_content = create_service_file(vars(args))
        if manage_service_file(args.name, service_content, "create") != 0:
            return 1

        run_systemctl("enable", f"{args.name}.service")
        run_systemctl("start", f"{args.name}.service")
        return 0
    else:
        service_content = create_service_file(vars(args))
        print(service_content)
        return 0


if __name__ == "__main__":
    sys.exit(service_handler())



import ast
import asyncio
import importlib
import importlib.util
import inspect
import json
import logging
import os
import re
import sys
import time
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from contextvars import copy_context
from functools import partial, wraps
from gzip import GzipFile
from itertools import chain
from logging.handlers import QueueHandler, QueueListener, TimedRotatingFileHandler
from multiprocessing import Queue as ProcessQueue
from pathlib import Path
from queue import Queue
from threading import Lock, RLock, Semaphore, Thread
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Optional,
    OrderedDict,
    Set,
    Tuple,
    Union,
)
from weakref import WeakSet

__all__ = [
    "lru_cache_ttl",
    "threads",
    "bg_task",
    "NamedLock",
    "FuncSchema",
    "InlinePB",
    "SizedTimedRotatingFileHandler",
    "get_type_default",
    "func_cmd",
    "file_import",
    "RotatingFileWriter",
    "get_function",
    "to_thread",
    "check_recursion",
]


def lru_cache_ttl(
    maxsize: int,
    ttl: Optional[Union[int, float]] = None,
    controls=False,
    auto_clear=True,
    timer=time.time,
):
    """A Least Recently Used (LRU) cache with a Time To Live (TTL) feature.

    Args:
        maxsize (int): maxsize of cache
        ttl (Optional[Union[int, float]], optional): time to live. Defaults to None.
        controls (bool, optional): set cache/ttl_clean attributes. Defaults to False.
        auto_clear (bool, optional): clear dead cache automatically. Defaults to True.
        timer (callable, optional): Defaults to time.time.

    Returns:
        callable: decorator function

    >>> import time
    >>> # test ttl
    >>> values = [1, 2]
    >>> @lru_cache_ttl(1, 0.1)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(1), time.sleep(0.11), func1(1)]
    [1, 1, None, 2]
    >>> # test maxsize
    >>> values = [1, 2, 3]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(2)(func)
    >>> [func1(i) for i in [1, 1, 1, 2, 2, 2, 3, 3, 3]]
    [1, 1, 1, 2, 2, 2, 3, 3, 3]
    >>> # test auto_clear=True, with controls
    >>> values = [1, 2, 3, 4]
    >>> func = lambda i: values.pop(0)
    >>> func1 = lru_cache_ttl(5, 0.1, controls=True, auto_clear=True)(func)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    1
    >>> # test auto_clear=False
    >>> values = [1, 2, 3, 4]
    >>> @lru_cache_ttl(5, 0.1, controls=True, auto_clear=False)
    ... def func1(i):
    ...     return values.pop(0)
    >>> [func1(1), func1(2), func1(3)]
    [1, 2, 3]
    >>> time.sleep(0.11)
    >>> func1(3)
    4
    >>> len(func1.cache)
    3
    """
    cache: OrderedDict = OrderedDict()
    move_to_end = cache.move_to_end
    popitem = cache.popitem
    next_clear_ts = timer()

    def ttl_clean(expire):
        for k, v in tuple(cache.items()):
            if v[1] <= expire:
                cache.pop(k, None)

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            nonlocal next_clear_ts
            now = timer()
            if ttl is None:
                expire_cleared = True
            else:
                expire_cleared = auto_clear and now > next_clear_ts
                if expire_cleared:
                    ttl_clean(now)
                    next_clear_ts = now + ttl
            # key = _make_key(args, kwargs, False)
            key = hash(tuple(chain(args, kwargs.items())))
            if key in cache:
                if expire_cleared is True or now < cache[key][1]:
                    # 1. key in cache; 2. cache is valid
                    # set newest, return result
                    move_to_end(key)
                    return cache[key][0]
            # call function
            result = func(*args, **kwargs)
            # ensure size
            while len(cache) >= maxsize:
                popitem(last=False)
            if ttl is None:
                # no need save expired ts
                cache[key] = (result,)
            else:
                # setitem(key, (result, now + ttl))
                cache[key] = (result, timer() + ttl)
            # {key: (result, expired_ts)}
            return result

        if controls:
            setattr(wrapped, "cache", cache)
            setattr(wrapped, "ttl_clean", ttl_clean)
        return wrapped

    return decorator


def threads(n: Optional[int] = None, executor_class=None, **kws):
    """Quickly convert synchronous functions to be concurrent. (similar to madisonmay/Tomorrow)

    >>> @threads(10)
    ... def test(i):
    ...     time.sleep(i)
    ...     return i
    >>> start = time.time()
    >>> tasks = [test(i) for i in [0.1] * 5]
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    5
    >>> for i in tasks:
    ...     i.result() if hasattr(i, 'result') else i
    0.1
    0.1
    0.1
    0.1
    0.1
    >>> time.time() - start < 0.2
    True
    >>> len(test.pool._threads)
    5
    >>> len(test.tasks)
    0
    >>> test.pool.shutdown()  # optional
    """
    pool = (executor_class or ThreadPoolExecutor)(max_workers=n, **kws)
    tasks: WeakSet = WeakSet()

    def decorator(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            future = pool.submit(func, *args, **kwargs)
            future.add_done_callback(lambda task: tasks.discard(task))
            tasks.add(future)
            return future

        setattr(wrapped, "pool", pool)
        setattr(wrapped, "tasks", tasks)
        return wrapped

    return decorator


_bg_tasks: Set[asyncio.Task] = set()


def bg_task(coro: Coroutine) -> asyncio.Task:
    """Avoid asyncio free-flying tasks, better to use the new asyncio.TaskGroup to avoid this in 3.11+. https://github.com/python/cpython/issues/91887

    Args:
        coro (Coroutine)

    Returns:
        _type_: Task

    """
    task = asyncio.create_task(coro)
    _bg_tasks.add(task)
    task.add_done_callback(_bg_tasks.discard)
    return task


LockType = Union[Lock, Semaphore, asyncio.Lock, asyncio.Semaphore]


class NamedLock:
    """Reusable named locks, support for timeouts, support for multiple concurrent locks.

    Demo::

        def test_named_lock():
            def test_sync():
                import time
                from concurrent.futures import ThreadPoolExecutor
                from threading import Lock, Semaphore

                def _test1():
                    with NamedLock("_test1", Lock, timeout=0.05) as lock:
                        time.sleep(0.2)
                        return bool(lock)

                with ThreadPoolExecutor(10) as pool:
                    tasks = [pool.submit(_test1) for _ in range(3)]
                    result = [i.result() for i in tasks]
                    assert result == [True, False, False], result
                assert len(NamedLock._SYNC_CACHE) == 1
                NamedLock.clear_unlocked()
                assert len(NamedLock._SYNC_CACHE) == 0

                def _test2():
                    with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                        time.sleep(0.2)
                        return bool(lock)

                with ThreadPoolExecutor(10) as pool:
                    tasks = [pool.submit(_test2) for _ in range(3)]
                    result = [i.result() for i in tasks]
                    assert result == [True, True, False], result

            def test_async():
                import asyncio

                async def main():
                    async def _test1():
                        async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                            await asyncio.sleep(0.2)
                            return bool(lock)

                    tasks = [asyncio.create_task(_test1()) for _ in range(3)]
                    result = [await i for i in tasks]
                    assert result == [True, False, False], result
                    assert len(NamedLock._ASYNC_CACHE) == 1
                    NamedLock.clear_unlocked()
                    assert len(NamedLock._ASYNC_CACHE) == 0

                    async def _test2():
                        async with NamedLock(
                            "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                        ) as lock:
                            await asyncio.sleep(0.2)
                            return bool(lock)

                    tasks = [asyncio.create_task(_test2()) for _ in range(3)]
                    result = [await i for i in tasks]
                    assert result == [True, True, False], result

                asyncio.run(main())

            test_sync()
            test_async()
    """

    _SYNC_CACHE: Dict[str, LockType] = {}
    _ASYNC_CACHE: Dict[str, LockType] = {}

    def __init__(self, name: str, default_factory: Callable, timeout=None):
        self.name = name
        self.default_factory = default_factory
        self.timeout = timeout
        self.lock = None

    @classmethod
    def clear_unlocked(cls):
        for cache in [cls._SYNC_CACHE, cls._ASYNC_CACHE]:
            for name, lock in list(cache.items()):
                if hasattr(lock, "locked") and not lock.locked():
                    cache.pop(name, None)
                elif isinstance(lock, Semaphore) and (
                    lock._value == 0 or not lock._cond._lock.locked()
                ):
                    cache.pop(name, None)

    def __enter__(self):
        if self.name in self._SYNC_CACHE:
            lock = self._SYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._SYNC_CACHE[self.name] = lock
        if lock.acquire(timeout=self.timeout):
            self.lock = lock
            return self
        else:
            return None

    def __exit__(self, *_):
        if self.lock:
            self.lock.release()

    async def __aenter__(self):
        if self.name in self._ASYNC_CACHE:
            lock = self._ASYNC_CACHE[self.name]
        else:
            lock = self.default_factory()
            self._ASYNC_CACHE[self.name] = lock
        try:
            await asyncio.wait_for(lock.acquire(), timeout=self.timeout)
            self.lock = lock
            return self
        except asyncio.TimeoutError:
            return None

    async def __aexit__(self, *_):
        if self.lock:
            self.lock.release()


class FuncSchema:
    """Parse the parameters and types required by a function into a dictionary, and convert an incoming parameter into the appropriate type.

    >>> def test(a, b: str, /, c=1, *, d=["d"], e=0.1, f={"f"}, g=(1, 2), h=True, i={1}, **kws):
    ...     return
    >>> FuncSchema.parse(test, strict=False)
    {'a': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}, 'b': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}, 'c': {'type': <class 'int'>, 'default': 1}, 'd': {'type': <class 'list'>, 'default': ['d']}, 'e': {'type': <class 'float'>, 'default': 0.1}, 'f': {'type': <class 'set'>, 'default': {'f'}}, 'g': {'type': <class 'tuple'>, 'default': (1, 2)}, 'h': {'type': <class 'bool'>, 'default': True}, 'i': {'type': <class 'set'>, 'default': {1}}, 'kws': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}}
    >>> def test(a):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `a` has no type and no default value.')
    >>> def test(b: str):
    ...     return
    >>> FuncSchema.parse(test, strict=True)
    {'b': {'type': <class 'str'>, 'default': <class 'inspect._empty'>}}
    >>> FuncSchema.parse(test, strict=True, fill_default=True)
    {'b': {'type': <class 'str'>, 'default': ''}}
    >>> def test(**kws):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `kws` has no type and no default value.')
    >>> def test(*args):
    ...     return
    >>> try:FuncSchema.parse(test, strict=True)
    ... except TypeError as e: e
    TypeError('Parameter `args` has no type and no default value.')
    >>> FuncSchema.convert("1", int)
    1
    >>> FuncSchema.convert("1", str)
    '1'
    >>> FuncSchema.convert("1", float)
    1.0
    >>> FuncSchema.convert(0, bool)
    False
    >>> FuncSchema.convert('1', bool)
    True
    >>> FuncSchema.convert('[[1, 1]]', dict)
    {1: 1}
    >>> FuncSchema.convert('{"1": "1"}', dict)
    {'1': '1'}
    >>> FuncSchema.convert('[1, 1]', set)
    {1}
    >>> FuncSchema.convert('[1, 1]', tuple)
    (1, 1)
    >>> FuncSchema.convert('[1, "1"]', list)
    [1, '1']
    >>> FuncSchema.to_string(1)
    '1'
    >>> FuncSchema.to_string("1")
    '1'
    >>> FuncSchema.to_string(1.0, float)
    '1.0'
    >>> FuncSchema.to_string(False)
    'false'
    >>> FuncSchema.to_string(True)
    'true'
    >>> FuncSchema.to_string({1: 1})
    '{"1": 1}'
    >>> FuncSchema.to_string({'1': '1'})
    '{"1": "1"}'
    >>> FuncSchema.to_string({1})
    '[1]'
    >>> FuncSchema.to_string((1, 1))
    '[1, 1]'
    >>> FuncSchema.to_string([1, '1'])
    '[1, "1"]'
    """

    ALLOW_TYPES = {int, float, str, tuple, list, set, dict, bool}
    JSON_TYPES = {tuple, list, set, dict, bool}

    @classmethod
    def parse(cls, function: Callable, strict=True, fill_default=False):
        sig = inspect.signature(function)
        result = {}
        for param in sig.parameters.values():
            if param.annotation is param.empty:
                if param.default is param.empty:
                    if strict:
                        raise TypeError(
                            f"Parameter `{param.name}` has no type and no default value."
                        )
                    else:
                        tp: Any = str
                else:
                    tp = type(param.default)
            else:
                tp = param.annotation
            if tp in cls.ALLOW_TYPES:
                default = param.default
                if fill_default and default is inspect._empty:
                    default = get_type_default(tp, default)
                result[param.name] = {"type": tp, "default": default}
        return result

    @classmethod
    def convert(cls, obj, target_type):
        if isinstance(obj, str) and target_type in cls.JSON_TYPES:
            return target_type(json.loads(obj))
        else:
            return target_type(obj)

    @classmethod
    def to_string(cls, obj, ensure_ascii=False):
        tp = type(obj)
        if isinstance(obj, str):
            return obj
        elif tp in cls.JSON_TYPES:
            if tp in {tuple, set}:
                obj = list(obj)
            return json.dumps(obj, ensure_ascii=ensure_ascii)
        elif tp in cls.ALLOW_TYPES:
            # {int, float}
            return str(obj)
        else:
            raise TypeError(f"Unsupported type: {tp}")


class InlinePB(object):
    """Inline progress bar.

    Demo::

        with InlinePB(100) as pb:
            for i in range(100):
                pb.add(1)
                time.sleep(0.03)
        # Progress:  41 / 100  41% [||||||         ] |   33 units/s
        with InlinePB(100) as pb:
            for i in range(1, 101):
                pb.update(i)
                time.sleep(0.03)
        # Progress:  45 / 100  45% [||||||         ] |   33 units/s

    """

    def __init__(
        self,
        total,
        start=0,
        maxlen=50,
        fresh_interval=0.1,
        timer=time.time,
        sig="|",
        sig_len=15,
    ):
        self.total = total
        self.done = start
        self.maxlen = maxlen
        self.cache = deque([], maxlen=maxlen)
        self.timer = timer
        self.last_fresh = self.timer()
        self.fresh_interval = fresh_interval
        self.sig = sig
        self.sig_len = sig_len

    def update(self, done):
        self.add(done - self.done)

    def add(self, num=1):
        self.done += num
        self.cache.append((self.done, self.timer()))
        if self.need_fresh():
            self.fresh()
            self.last_fresh = self.timer()

    def need_fresh(self):
        if self.timer() - self.last_fresh > self.fresh_interval:
            return True
        else:
            return False

    def speed(self) -> int:
        if len(self.cache) > 1:
            a, b = self.cache[0], self.cache[-1]
            return round((b[0] - a[0]) / (b[1] - a[1]))
        elif self.cache:
            return round(self.done / (self.timer() - self.cache[0][1]))
        else:
            return 0

    def __enter__(self):
        self._fill = len(str(self.total))
        self._end = f"{' ' * 10}\r"
        return self

    def __exit__(self, *_):
        if not any(_):
            self.fresh()
            print(flush=True)

    def sig_string(self, percent):
        return f"[{self.sig * int(self.sig_len * percent / 100)}{' ' * (self.sig_len - int(self.sig_len * percent / 100))}]"

    def fresh(self):
        percent = int(100 * self.done / self.total)
        done = f"{self.done}".rjust(self._fill, " ")
        print(
            f"Progress: {done} / {self.total} {percent: >3}% {self.sig_string(percent)} | {self.speed(): >4} units/s",
            end=self._end,
            flush=True,
        )


class SizedTimedRotatingFileHandler(TimedRotatingFileHandler):
    """TimedRotatingFileHandler with maxSize, to avoid files that are too large.


    Demo::

        import logging
        import time
        from morebuiltins.funcs import SizedTimedRotatingFileHandler

        logger = logging.getLogger("test1")
        h = SizedTimedRotatingFileHandler(
            "logs/test1.log", "d", 1, 3, maxBytes=1, ensure_dir=True
        )
        h.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
        logger.addHandler(h)

        for i in range(5):
            logger.warning(str(i) * 102400)
            time.sleep(1)
        # 102434 test1.log
        # 102434 test1.log.20241113_231000
        # 102434 test1.log.20241113_231001
        # 102434 test1.log.20241113_231002
        logger = logging.getLogger("test2")
        h = SizedTimedRotatingFileHandler(
            "logs/test2.log", "d", 1, 3, maxBytes=1, compress=True
        )
        h.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
        logger.addHandler(h)

        for i in range(5):
            logger.warning(str(i) * 102400)
            time.sleep(1)
        # 102434 test2.log
        #    186 test2.log.20241113_231005.gz
        #    186 test2.log.20241113_231006.gz
        #    186 test2.log.20241113_231007.gz

    """

    do_compress_delay = 0.1

    def __init__(
        self,
        filename,
        when="h",
        interval=1,
        backupCount=0,
        maxBytes=0,
        encoding=None,
        delay=False,
        utc=False,
        compress=False,
        ensure_dir=True,
    ):
        """
        Initialize the timed backup file handler.

        :param filename: The name of the log file.
        :param when: The time unit for timed backups, can be "h" (hours) or "d" (days)
        :param interval: The interval for timed backups, with the unit determined by the 'when' parameter
        :param backupCount: The maximum number of backup files to keep
        :param maxBytes: The file size limit before triggering a backup (0 means no limit)
        :param encoding: The encoding of the file
        :param delay: Whether to delay opening the file until the first write
        :param utc: Whether to use UTC time for naming backups
        """
        self.log_path = Path(filename)
        if ensure_dir:
            self.log_path.parent.mkdir(parents=True, exist_ok=True)
        super().__init__(filename, when, interval, backupCount, encoding, delay, utc)
        self.maxBytes = maxBytes
        self.suffix = "%Y%m%d_%H%M%S"
        self.compress = compress
        self.need_compress = False
        self.comress_chunk_size = 64 * 1024
        self.extMatch = re.compile(r"^\d{4}\d{2}\d{2}_\d{2}\d{2}\d{2}$", re.ASCII)

    def do_compress_async(self):
        if getattr(self, "_background_compress", None):
            return
        self._compressing = Thread(
            target=self.do_compress, daemon=False, name="do_compress_async"
        )
        self._compressing.start()

    def gzip_log(self, path: Path):
        try:
            temp_path = path.with_suffix(".tmp.gz")
            with open(path, "rb") as f:
                with GzipFile(temp_path, "wb") as gz:
                    size = 0
                    lines = []
                    for line in f:
                        size += len(line)
                        lines.append(line)
                        if size > self.comress_chunk_size:
                            gz.writelines(lines)
                            size = 0
                            lines = []
                    if lines:
                        gz.writelines(lines)
            if path.is_file():
                try:
                    path.unlink()
                    temp_path.rename(path.with_suffix(f"{path.suffix}.gz"))
                except OSError:
                    pass
        except OSError:
            pass
        finally:
            try:
                temp_path.unlink(missing_ok=True)
            except OSError:
                pass

    def do_compress(self):
        try:
            while self.need_compress:
                self.need_compress = False
                time.sleep(self.do_compress_delay)
                now_suffix = f".{time.strftime(self.suffix)}"
                path_list = []
                for path in self.log_path.parent.glob(f"{self.log_path.name}.*"):
                    if path.suffix == now_suffix:
                        self.need_compress = True
                        continue
                    for time_suffix in path.name.split(".")[-2:]:
                        if self.extMatch.match(time_suffix):
                            path_list.append((time_suffix, path))
                path_list.sort()
                target_index = len(path_list) - self.backupCount
                for index, (_, path) in enumerate(path_list):
                    if index < target_index:
                        try:
                            path.unlink(missing_ok=True)
                        except OSError:
                            pass
                        continue
                    elif path.suffix == ".gz":
                        continue
                    else:
                        self.gzip_log(path)
        finally:
            self._compressing = None

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.
        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if super().shouldRollover(record):
            return True

        if self.maxBytes > 0:
            if self.stream.tell() >= self.maxBytes:
                return True
        return False

    def doRollover(self):
        """
        Do a rollover, as described by the base class documentation.
        However, also check for the maxBytes parameter and rollover if needed.
        """
        try:
            super().doRollover()
        except OSError:
            pass
        finally:
            self.need_compress = True
        if self.compress:
            self.do_compress_async()

    def getFilesToDelete(self):
        # always return null list
        if self.compress:
            return []
        else:
            return super().getFilesToDelete()

    def __del__(self):
        if self.compress:
            self.do_compress()


def get_type_default(tp, default=None):
    """Get the default value for a type. {int: 0, float: 0.0, bytes: b"", str: "", list: [], tuple: (), set: set(), dict: {}}"""
    return {
        int: 0,
        float: 0.0,
        bytes: b"",
        str: "",
        list: [],
        tuple: (),
        set: set(),
        dict: {},
    }.get(tp, default)


def func_cmd(function: Callable, run=True, auto_default=False):
    """Handle function with argparse, typing-hint is nessessary.

    Demo::

        def test(str: str, /, int=1, *, list=["d"], float=0.1, set={"f"}, tuple=(1, 2), bool=True, dict={"k": 1}):
            \"\"\"Test demo function.

            Args:
                str (str): str.
                int (int, optional): int. Defaults to 1.
                list (list, optional): list. Defaults to ["d"].
                float (float, optional): float. Defaults to 0.1.
                set (dict, optional): set. Defaults to {"f"}.
                tuple (tuple, optional): tuple. Defaults to (1, 2).
                bool (bool, optional): bool. Defaults to True.
                dict (dict, optional): dict. Defaults to {"k": 1}.
            \"\"\"
            print(locals())

        # raise ValueError if auto_default is False and user do not input nessessary args.
        func_cmd(test, auto_default=False)

        CMD args:

        > python app.py
        ValueError: `str` has no default value.

        > python app.py --str 1 --int 2 --float 1.0 --list "[1,\"a\"]" --tuple "[2,\"b\"]" --set "[1,1,2]" --dict "{\"k\":\"v\"}"
        {'str': '1', 'int': 2, 'list': [1, 'a'], 'float': 1.0, 'set': {1, 2}, 'tuple': (2, 'b'), 'bool': True, 'dict': {'k': 'v'}}

        > python app.py -s 1 -i 2 -f 1.0 -l "[1,\"a\"]" -t "[2,\"b\"]" -s "[1,1,2]" -d "{\"k\":\"v\"}"
        {'str': '[1,1,2]', 'int': 2, 'list': [1, 'a'], 'float': 1.0, 'set': {'f'}, 'tuple': (2, 'b'), 'bool': True, 'dict': {'k': 'v'}}

        > python app.py -h
        usage: Test demo function.

            Args:
                str (str): str.
                int (int, optional): int. Defaults to 1.
                list (list, optional): list. Defaults to ["d"].
                float (float, optional): float. Defaults to 0.1.
                set (dict, optional): set. Defaults to {"f"}.
                tuple (tuple, optional): tuple. Defaults to (1, 2).
                bool (bool, optional): bool. Defaults to True.
                dict (dict, optional): dict. Defaults to {"k": 1}.


        options:
        -h, --help            show this help message and exit
        -s STR, --str STR     {'type': <class 'str'>, 'default': <class 'inspect._empty'>}
        -i INT, --int INT     {'type': <class 'int'>, 'default': 1}
        -l LIST, --list LIST  {'type': <class 'list'>, 'default': ['d']}
        -f FLOAT, --float FLOAT
                                {'type': <class 'float'>, 'default': 0.1}
        -se SET, --set SET    {'type': <class 'set'>, 'default': {'f'}}
        -t TUPLE, --tuple TUPLE
                                {'type': <class 'tuple'>, 'default': (1, 2)}
        -b BOOL, --bool BOOL  {'type': <class 'bool'>, 'default': True}
        -d DICT, --dict DICT  {'type': <class 'dict'>, 'default': {'k': 1}}
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(usage=function.__doc__)
    schema = FuncSchema.parse(function, strict=True, fill_default=False)
    seen_shorten = {"h"}
    for key, value in schema.items():
        args = [f"--{key}"]
        for i in range(1, 4):
            if i > len(key):
                break
            short = key[:i]
            if short not in seen_shorten:
                seen_shorten.add(short)
                args.insert(0, f"-{short}")
                break
        kwargs = dict(dest=key, help=str(value))
        if value["type"] in {int, float, str}:
            kwargs["type"] = value["type"]
        else:
            kwargs["type"] = str
        if value["default"] is inspect._empty:
            kwargs["default"] = get_type_default(value["type"])
        else:
            kwargs["default"] = value["default"]
        parser.add_argument(*args, **kwargs)
    parsed = parser.parse_args()
    kwargs = {}
    args = []
    for key, value in schema.items():
        if value["default"] is inspect._empty:
            if auto_default:
                args.append(FuncSchema.convert(getattr(parsed, key), value["type"]))
            else:
                raise ValueError(f"`{key}` has no default value.")
        else:
            kwargs[key] = FuncSchema.convert(getattr(parsed, key), value["type"])
    if run:
        return function(*args, **kwargs)
    else:
        return args, kwargs


def file_import(file_path, names):
    """Import function from file path.

    Demo::
        >>> from pathlib import Path
        >>> file_path = Path(__file__).parent / "utils.py"
        >>> list(file_import(file_path, ["get_hash", "find_jsons"]).keys())
        ['get_hash', 'find_jsons']
    """
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return {name: getattr(module, name) for name in names}


class RotatingFileWriter:
    """RotatingFileWriter class for writing to a file with rotation support.

    Demo::

        >>> # test normal usage
        >>> writer = RotatingFileWriter("test.log", max_size=10 * 1024, max_backups=1)
        >>> writer.write("1" * 10)
        >>> writer.path.stat().st_size
        0
        >>> writer.flush()
        >>> writer.path.stat().st_size
        10
        >>> writer.clean_backups(writer.max_backups)
        >>> writer.unlink_file()
        >>> # test rotating
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=2)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15, flush=True)
        >>> writer.path.stat().st_size
        15
        >>> len(writer.backup_path_list())
        2
        >>> writer.clean_backups(writer.max_backups)
        >>> writer.unlink_file()
        >>> # test no backups
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=0)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15)
        >>> writer.write("1" * 15, flush=True)
        >>> writer.path.stat().st_size
        15
        >>> len(writer.backup_path_list())
        0
        >>> writer.clean_backups(writer.max_backups)
        >>> len(writer.backup_path_list())
        0
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=3)
        >>> writer.print("1" * 100)
        >>> writer.unlink(rotate=False)
        >>> len(writer.backup_path_list())
        1
        >>> writer.unlink(rotate=True)
        >>> len(writer.backup_path_list())
        0
        >>> writer = RotatingFileWriter("test.log", max_size=20, max_backups=3, compress=True)
        >>> writer.print("1" * 100)
        >>> len(writer.backup_path_list())
        1
        >>> writer.unlink(rotate=True)
        >>> len(writer.backup_path_list())
        0
    """

    check_exist_every = 100

    def __init__(
        self,
        path: Union[Path, str],
        max_size=5 * 1024**2,
        max_backups=0,
        encoding="utf-8",
        errors=None,
        buffering=-1,
        newline=None,
        compress=False,
    ):
        if max_backups < 0:
            raise ValueError("max_backups must be greater than -1, 0 for itself.")
        self._compress_threads: WeakSet = WeakSet()
        self._rotate_lock = RLock()
        self.path = Path(path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self.max_size = max_size
        self.max_backups = max_backups
        self.encoding = encoding
        self.errors = errors
        self.buffering = buffering
        self.newline = newline
        self.compress = compress
        self.file = self.reopen_file()
        self._check_exist_count = self.check_exist_every + 1

    def get_suffix(self):
        return time.strftime("%Y%m%d%H%M%S")

    def unlink_file(self):
        return self.unlink(rotate=False)

    def unlink(self, rotate=True, parent=False):
        self.close_file()
        self.path.unlink(missing_ok=True)
        if rotate:
            self.clean_backups(count=self.max_backups + 1)
        if parent:
            for _ in self.path.parent.iterdir():
                return
            else:
                self.path.parent.rmdir()

    def close(self):
        return self.close_file()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.shutdown()

    def shutdown(self):
        self.close_file()
        if self.compress:
            for t in self._compress_threads:
                t.join()

    def close_file(self):
        file_obj = getattr(self, "file", None)
        if file_obj and not file_obj.closed:
            file_obj.close()
            self.file = None

    def reopen_file(self):
        self.close_file()
        self.file = self.path.open(
            "a",
            encoding=self.encoding,
            errors=self.errors,
            buffering=self.buffering,
            newline=self.newline,
        )
        return self.file

    def check_exist(self):
        return not (
            self._check_exist_count > self.check_exist_every and not self.path.is_file()
        )

    def rotate(self, new_length):
        with self._rotate_lock:
            if self.need_rotate(new_length):
                if self.max_backups > 0:
                    self.close_file()
                    _suffix = self.get_suffix()
                    for index in range(self.max_backups):
                        if index:
                            suffix = f"{_suffix}_{index}"
                        else:
                            suffix = _suffix
                        target_path = self.path.with_name(f"{self.path.name}.{suffix}")
                        if target_path.is_file():
                            # already rotated
                            continue
                        else:
                            break
                    else:
                        raise RuntimeError(
                            "max_backups is too small for writing too fast"
                        )
                    self.path.rename(target_path)
                    self.reopen_file()
                    if not self.compress:
                        self.clean_backups(count=None)
                else:
                    self.file.seek(0)
                    self.file.truncate()

    def do_compress(self):
        with self._rotate_lock:
            for path in self.path.parent.glob(f"{self.path.name}.*"):
                if path.name == self.path.name:
                    continue
                elif path.suffix == ".gz":
                    continue
                temp_path = path.with_name(path.name + ".gz")
                with GzipFile(temp_path, "wb") as gzip_file:
                    with path.open("rb") as src_file:
                        for line in src_file:
                            gzip_file.write(line)
                path.unlink(missing_ok=True)
            self.clean_backups()

    def need_rotate(self, new_length):
        return self.max_size and self.file.tell() + new_length > self.max_size

    def ensure_file(self, new_length=0):
        if not self.file:
            self.reopen_file()
        elif not self.check_exist():
            self.reopen_file()
        elif self.need_rotate(new_length):
            self.rotate(new_length)
            if self.compress:
                t = Thread(target=self.do_compress)
                t.start()
                self._compress_threads.add(t)

    def backup_path_list(self):
        return list(self.path.parent.glob(f"{self.path.name}.*"))

    def clean_backups(self, count=None):
        """Clean oldest {count} backups, if count is None, it will clean up to max_backups."""
        with self._rotate_lock:
            path_list = self.backup_path_list()
            if path_list:
                if count is None:
                    count = len(path_list) - self.max_backups
                if count > 0:
                    path_list.sort(key=lambda x: x.stat().st_mtime)
                    for deleted, path in enumerate(path_list, 1):
                        path.unlink(missing_ok=True)
                        if deleted >= count:
                            break

    def flush(self):
        self.file.flush()

    def write(self, text: str, flush=False):
        self._check_exist_count += 1
        self.ensure_file(len(text))
        self.file.write(text)
        if flush:
            self.file.flush()

    def print(self, *strings, end="\n", sep=" ", flush=False):
        text = f"{sep.join(map(str, strings))}{end}"
        self.write(text, flush=flush)

    def __del__(self):
        self.shutdown()


def get_function(entrypoint: str):
    """Get the function object from entrypoint.

    Demo::

        >>> get_function("urllib.parse:urlparse").__name__
        'urlparse'
    """
    module, _, function = entrypoint.partition(":")
    return getattr(importlib.import_module(module), function)


async def to_thread(func, /, *args, **kwargs):
    """Asynchronously run function *func* in a separate thread, same as `asyncio.to_thread` in python 3.9+."""
    func_call = partial(copy_context().run, func, *args, **kwargs)
    return await asyncio.get_running_loop().run_in_executor(None, func_call)


class AsyncQueueListener(QueueListener):
    """Asynchronous non-blocking QueueListener that manages logger handlers.
    logger is a logging.Logger instance.
    queue is a Queue or ProcessQueue instance.
    respect_handler_level is a boolean that determines if the handler level should be respected.

    Example:

        async def main():
            # Create logger with a blocking handler
            logger = logging.getLogger("example")
            logger.setLevel(logging.INFO)
            stream_handler = logging.StreamHandler()
            stream_handler.setFormatter(
                logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
            )
            logger.addHandler(stream_handler)
            # Use async queue listener
            async with AsyncQueueListener(logger):
                # Log won't block the event loop
                for i in range(5):
                    logger.info("log info")
                    logger.debug("log debug")
                    await asyncio.sleep(0.01)
    """

    def __init__(
        self,
        logger: logging.Logger,
        queue: Optional[Union[Queue, ProcessQueue]] = None,
        respect_handler_level=True,
    ):
        self.logger = logger
        self.queue = queue or Queue()
        # Store original handlers
        self.original_handlers = list(logger.handlers)
        # Get handlers that might block
        self.blocking_handlers = [
            h for h in self.original_handlers if not isinstance(h, QueueHandler)
        ]
        # Initialize parent with queue and blocking handlers
        super().__init__(
            self.queue,
            *self.blocking_handlers,
            respect_handler_level=respect_handler_level,
        )

    def _switch_to_queue_handler(self):
        """Switch handlers in a blocking context"""
        # Remove original handlers
        for handler in self.original_handlers:
            self.logger.removeHandler(handler)
        # Add queue handler
        self.logger.addHandler(QueueHandler(self.queue))

    def _restore_original_handlers(self):
        """Restore original handlers in a blocking context"""
        # Remove queue handler
        self.logger.removeHandler(self.logger.handlers[0])
        # Restore original handlers
        for handler in self.original_handlers:
            self.logger.addHandler(handler)

    def stop(self):
        super().stop()
        self._restore_original_handlers()

    def start(self):
        self._switch_to_queue_handler()
        super().start()

    async def __aenter__(self):
        await to_thread(self.start)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await to_thread(self.stop)


# alias for AsyncQueueListener
async_logger = AsyncQueueListener


def check_recursion(function: Callable, return_error=False):
    """Check if a function is recursive by inspecting its AST.
    Returns True if the function calls itself, otherwise False.

    Demo::
        >>> def recursive_func():
        ...     return recursive_func()
        >>> check_recursion(recursive_func)
        True
        >>> def non_recursive_func():
        ...     return 1 + 1
        >>> check_recursion(non_recursive_func)
        False
        >>> # print is a std-lib function
        >>> check_recursion(print, return_error=False)
        >>> type(check_recursion(print, return_error=True))
        <class 'TypeError'>
    """
    try:
        source = inspect.getsource(function)
        tree = ast.parse(source)
        func_name = function.__name__
        for node in ast.walk(tree):
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Name)
                and node.func.id == func_name
            ):
                return True
        return False
    except Exception as e:
        if return_error:
            return e
        else:
            return None


class LineProfiler:
    """Line-by-line performance profiler."""

    stdout = sys.stdout

    def __init__(self):
        self.line_times: List[Tuple[int, float, str]] = []
        self.start_time = 0.0
        self.total_time = 0.0
        self.line_cache: Dict[str, list] = {}

    def trace_calls(self, frame, event, arg):
        """Trace each line of function calls"""
        if event == "line":
            current_time = time.perf_counter()
            line_no = frame.f_lineno
            filename = frame.f_code.co_filename
            # Get the code content of current line
            try:
                if filename in self.line_cache:
                    lines = self.line_cache[filename]
                else:
                    with open(filename, "r", encoding="utf-8", errors="replace") as f:
                        lines = f.readlines()
                    self.line_cache[filename] = lines
                if lines and line_no <= len(lines):
                    line_content = lines[
                        line_no - 1
                    ].rstrip()  # Only remove right spaces, keep left indentation
                    # Store original code and filename for later base indentation calculation
                    if not hasattr(self, "source_lines"):
                        self.source_lines = lines
                        self.target_filename = filename
                else:
                    line_content = "-"
            except Exception:
                line_content = "-"
                self.line_cache.setdefault(filename, [])
            self.line_times.append((line_no, current_time, line_content))
        return self.trace_calls

    def _get_function_base_indent(self, func_name: str):
        """Get the base indentation level of function definition"""
        if not hasattr(self, "source_lines"):
            return 0
        # Find function definition line
        for line in self.source_lines:
            if f"def {func_name}(" in line:
                # Calculate indentation level of function definition
                return len(line) - len(line.lstrip())
        return 0

    def calculate_and_print_stats(self, func_name: str):
        """Calculate and print statistics for each line"""
        if len(self.line_times) < 2:
            return
        # Calculate base indentation level of function
        base_indent = self._get_function_base_indent(func_name)
        print(f"`{func_name}` profiling report:", file=self.stdout)
        start_time = time.strftime("%Y-%m-%d %H:%M:%S")
        print(
            f"{start_time} | Total: {self.total_time * 1000:.3f} ms",
            file=self.stdout,
        )
        # Calculate execution time for each line
        line_durations = []
        for i in range(1, len(self.line_times)):
            prev_time = self.line_times[i - 1][1]
            curr_time = self.line_times[i][1]
            duration = curr_time - prev_time
            line_durations.append(
                (
                    self.line_times[i - 1][0],  # line number
                    duration,  # duration
                    self.line_times[i - 1][2],  # code content
                )
            )
        # Sort by line number and merge time for same lines
        line_stats: Dict[int, Dict[str, Any]] = {}
        for line_no, duration, code in line_durations:
            if line_no not in line_stats:
                line_stats[line_no] = {
                    "total_time": 0,
                    "count": 0,
                    "code": code,
                    "timestamps": [],
                }
            line_stats[line_no]["total_time"] += duration
            line_stats[line_no]["count"] += 1
        # Print detailed information for each line
        print(f"{'=' * 95}", file=self.stdout)
        print(
            f"{'Line':>6} {'%':>4} {'Total(ms)':>12} {'Count':>8} {'Avg(ms)':>12}  {'Source Code':<40}",
            file=self.stdout,
        )
        print(f"{'-' * 95}", file=self.stdout)
        sorted_lines = sorted(line_stats.items())
        for line_no, stats in sorted_lines:
            total_time_ms = stats["total_time"] * 1000  # convert to milliseconds
            count = stats["count"]
            avg_time_ms = total_time_ms / count if count > 0 else 0
            percentage = int(
                (stats["total_time"] / self.total_time * 100)
                if self.total_time > 0
                else 0
            )  # convert to integer
            # Handle code indentation display
            code = stats["code"]
            # Get current line indentation
            current_indent = len(code) - len(code.lstrip())
            # First level indentation under def (usually 4 spaces)
            first_level_indent = base_indent + 4
            # Only keep spaces relative to first level indentation
            relative_indent = max(0, current_indent - first_level_indent)
            indent_spaces = " " * relative_indent
            code_with_relative_indent = indent_spaces + code.lstrip()
            code_preview = (
                code_with_relative_indent[:38] + ".."
                if len(code_with_relative_indent) > 40
                else code_with_relative_indent
            )
            print(
                f"{line_no:>6} {percentage:>4} {total_time_ms:>12.3f} {count:>8} {avg_time_ms:>12.3f}  {code_preview:<40}",
                file=self.stdout,
            )
        print(f"{'=' * 95}", file=self.stdout, flush=True)


def line_profiler(func: Callable) -> Callable:
    """Decorator to profile a function line-by-line.

    Demo usage:
        >>> import sys, io
        >>> LineProfiler.stdout = io.StringIO()  # Redirect stdout to capture print output
        >>> @line_profiler
        ... def example_function():
        ...     result = 0
        ...     for i in range(10):
        ...         result += i  # Simulate some work
        ...     return result
        >>> example_function()
        45
        >>> output = LineProfiler.stdout.getvalue()
        >>> output.splitlines()[0].startswith("`example_function` profiling report:")
        True
        >>> LineProfiler.stdout = sys.stdout  # Restore original stdout

    # `example_function` profiling report:
    # 2025-07-26 17:09:58 | Total: 1.122 ms
    # ===============================================================================================
    #   Line    %    Total(ms)    Count      Avg(ms)  Source Code
    # -----------------------------------------------------------------------------------------------
    #      3   73        0.825        1        0.825  -
    #      4    3        0.040       11        0.004  -
    #      5    4        0.050       10        0.005  -
    # ===============================================================================================
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        profiler = LineProfiler()
        # Record start time
        start_time = time.perf_counter()
        profiler.start_time = start_time
        # Set tracer
        old_trace = sys.gettrace()
        sys.settrace(profiler.trace_calls)
        try:
            # Execute function
            result = func(*args, **kwargs)
        finally:
            # Restore original tracer
            sys.settrace(old_trace)
            # Record end time
            end_time = time.perf_counter()
            profiler.total_time = end_time - start_time
            # Calculate and print statistics
            profiler.calculate_and_print_stats(func.__name__)
        return result

    return wrapper


def test_bg_task():
    async def _test_bg_task():
        async def coro():
            return True

        task = bg_task(coro())
        assert await task is True
        result = (task.done(), len(_bg_tasks))
        assert result == (True, 0), result

    asyncio.run(_test_bg_task())


def test_named_lock():
    def test_sync():
        import time
        from concurrent.futures import ThreadPoolExecutor
        from threading import Lock, Semaphore

        def _test1():
            with NamedLock("_test1", Lock, timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test1) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, False, False], result
        assert len(NamedLock._SYNC_CACHE) == 1
        NamedLock.clear_unlocked()
        assert len(NamedLock._SYNC_CACHE) == 0

        def _test2():
            with NamedLock("_test2", lambda: Semaphore(2), timeout=0.05) as lock:
                time.sleep(0.2)
                return bool(lock)

        with ThreadPoolExecutor(10) as pool:
            tasks = [pool.submit(_test2) for _ in range(3)]
            result = [i.result() for i in tasks]
            assert result == [True, True, False], result

    def test_async():
        import asyncio

        async def main():
            async def _test1():
                async with NamedLock("_test1", asyncio.Lock, timeout=0.05) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test1()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, False, False], result
            assert len(NamedLock._ASYNC_CACHE) == 1
            NamedLock.clear_unlocked()
            assert len(NamedLock._ASYNC_CACHE) == 0

            async def _test2():
                async with NamedLock(
                    "_test2", lambda: asyncio.Semaphore(2), timeout=0.05
                ) as lock:
                    await asyncio.sleep(0.2)
                    return bool(lock)

            tasks = [asyncio.create_task(_test2()) for _ in range(3)]
            result = [await i for i in tasks]
            assert result == [True, True, False], result

        asyncio.run(main())

    test_sync()
    test_async()


def test_AsyncQueueListener():
    """Test AsyncQueueListener logs without blocking"""

    async def _test():
        from io import StringIO

        mock_stdout = StringIO()
        # Create logger with a blocking handler
        logger = logging.getLogger("example")
        logger.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler(stream=mock_stdout)
        stream_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        )
        logger.addHandler(stream_handler)
        # Use async queue listener
        async with AsyncQueueListener(logger):
            # Log won't block the event loop
            for i in range(5):
                logger.info("log info")
                logger.debug("log debug")
        text = mock_stdout.getvalue()
        assert text.count("log info") == 5
        assert text.count("log debug") == 0

    asyncio.run(_test()) is True


def test_utils():
    test_bg_task()
    test_named_lock()
    test_AsyncQueueListener()


def test():
    global __name__
    __name__ = "morebuiltins.funcs"
    import doctest

    doctest.testmod()
    test_utils()


if __name__ == "__main__":
    test()



import atexit
import os
import time
import typing
from contextlib import closing

from multiprocessing.shared_memory import SharedMemory

__all__ = ["PLock", "SharedBytes"]


class PLock:
    """A simple process lock using shared memory, for singleton control.
    Use `with` context or `close_atexit` to ensure the shared memory is closed in case the process crashes.

    Args:
        name (str): name of the shared memory
        force (bool, optional): whether to force rewrite the existing shared memory. Defaults to False.
        close_atexit (bool, optional): whether to close the shared memory at process exit. Defaults to False, to use __del__ or __exit__ instead.

    Demo:

        >>> test_pid = 123456 # test pid, often set to None for current process
        >>> plock = PLock("test_lock", force=False, close_atexit=True, pid=test_pid)
        >>> plock.locked
        True
        >>> try:
        ...     plock2 = PLock("test_lock", force=False, close_atexit=True, pid=test_pid + 1)
        ...     raise RuntimeError("Should not be here")
        ... except RuntimeError:
        ...     True
        True
        >>> plock3 = PLock("test_lock", force=True, close_atexit=True, pid=test_pid + 1)
        >>> plock3.locked
        True
        >>> plock.locked
        False
        >>> PLock.wait_for_free(name="test_lock", timeout=0.1, interval=0.01)
        False
        >>> plock.close()
        >>> plock3.close()
        >>> PLock.wait_for_free(name="test_lock", timeout=0.1, interval=0.01)
        True
    """

    DEFAULT_SIZE = 4  # 4 bytes, means 2^32 = 4GB
    DEFAULT_BYTEORDER = "little"

    def __init__(self, name: str, force=False, close_atexit=False, pid=None):
        self.name = name
        self.pid = pid or os.getpid()
        self.force = force
        self.shm = None
        # whether the shared memory is closed
        self._closed = False
        if close_atexit:
            atexit.register(self.close)
        self.init()
        if self.shm is None:
            raise RuntimeError(f"Failed to create shared memory {name}")

    @staticmethod
    def wait_for_free(name: str, timeout=3, interval=0.1):
        """Wait for the shared memory to be free."""
        start = time.time()
        while time.time() - start < timeout:
            try:
                SharedMemory(name=name).close()
                time.sleep(interval)
            except FileNotFoundError:
                return True
        return False

    def init(self):
        if self._closed:
            raise RuntimeError("Already closed")
        ok = True
        try:
            self.shm = SharedMemory(name=self.name, create=True, size=self.DEFAULT_SIZE)
        except FileExistsError:
            self.shm = SharedMemory(name=self.name)
            if not self.force:
                ok = False
        if not ok:
            mem_pid = self.mem_pid
            self.close()
            raise RuntimeError(
                f"Locked by another process: {mem_pid} != {self.pid}(self)"
            )
        self.set_mem_pid()
        if not self.locked:
            raise ValueError(
                f"Failed to write PID to shared memory. {self.mem_pid} != {self.pid}(self)"
            )

    def set_mem_pid(self, pid=None):
        if pid is None:
            pid = self.pid
        self.buf[: self.DEFAULT_SIZE] = pid.to_bytes(
            self.DEFAULT_SIZE, byteorder=self.DEFAULT_BYTEORDER
        )

    @property
    def buf(self):
        if self.shm is None:
            raise RuntimeError("Shared memory is not initialized")
        return self.shm.buf

    def get_mem_pid(self):
        return int.from_bytes(
            self.buf[: self.DEFAULT_SIZE], byteorder=self.DEFAULT_BYTEORDER
        )

    @property
    def mem_pid(self):
        return self.get_mem_pid()

    @property
    def locked(self):
        return self.mem_pid == self.pid

    def close(self):
        if self._closed:
            return
        self._closed = True
        if self.shm:
            try:
                locked = self.locked
                self.shm.close()
                if locked:
                    # only unlink if self.pid is the owner of the shared memory
                    self.shm.unlink()
            except Exception:
                pass

    def __enter__(self):
        return self

    def __exit__(self, *_, **_kwargs):
        self.close()

    def __del__(self):
        self.close()


class SharedBytes:
    """Shared Memory for Python, for python 3.8+.
    This module provides a simple way to create and manage shared memory segments, shared between different processes.
    Shared memory is faster than other IPC methods like pipes or queues, and it allows for direct access to the memory.

    Demo:

    >>> sb = SharedBytes(name="test", data=b"Hello, World!", unlink_on_exit=True)
    >>> # The size of the shared memory is 18 bytes (5 bytes for header + 13 bytes for data), but mac os may return more than 18 bytes.
    >>> sb.size > 10
    True
    >>> sb.get(name="test")
    b'Hello, World!'
    >>> sb.re_create(b"New Data")
    >>> sb.get(name="test")
    b'New Data'
    >>> sb.close()
    >>> sb.get(name="test", default=b"")  # This will raise ValueError since the shared memory is closed
    b''
    """

    closed: bool = True
    # max_size: 2 ** (i * 8). 1: 256 B, 2: 64 KB, 3: 16 MB, 4: 4 GB, 5: 1 TB, 6: 256 TB, 7: 64 PB, 8: 16 EB, defaults to 5(1TB).
    head_length: int = 5
    byteorder: typing.Literal["little", "big"] = "little"

    def __init__(self, name: str, data: bytes, head_length=None, unlink_on_exit=False):
        self.name = name
        self.head_length = head_length or self.head_length
        if unlink_on_exit:
            atexit.register(self.close)
        self.create(data)

    @property
    def size(self) -> int:
        if self.closed:
            raise ValueError("Shared memory is closed")
        return self.shm.size

    def create(self, data: bytes):
        if not self.closed:
            raise ValueError("Shared memory is already created, please use re_create")
        size = len(data)
        head_length = self.head_length
        if size > 2 ** (head_length * 8):
            right_head_length = 0
            for i in range(head_length, 15):
                if size < 2 ** (i * 8):
                    right_head_length = i
                    break
            raise ValueError(
                f"data size {size} is too large, max size is {2 ** (head_length * 8)}, raise head_length at least {right_head_length}"
            )
        total_size = head_length + len(data)
        self.shm = SharedMemory(name=self.name, create=True, size=total_size)
        head = size.to_bytes(head_length, byteorder=self.byteorder)
        self.buf[:total_size] = head + data
        self.closed = False

    @property
    def buf(self):
        if self.shm is None:
            raise RuntimeError("Shared memory is not initialized")
        return self.shm.buf

    def re_create(self, data: bytes):
        if self.closed:
            raise ValueError("Shared memory is closed")
        self.close()
        self.create(data)

    def close(self):
        if not self.closed:
            self.shm.close()
            self.shm.unlink()
            self.closed = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @classmethod
    def get(cls, name: str, head_length=None, default=...) -> bytes:
        try:
            head_length = head_length or cls.head_length
            with closing(SharedMemory(name=name)) as shm:
                head = bytes(shm.buf[:head_length])
                body_length = int.from_bytes(head, byteorder=cls.byteorder)
                data = bytes(shm.buf[head_length : head_length + body_length])
                return data
        except FileNotFoundError:
            if default is ...:
                raise KeyError(f"Shared memory {name} not found")
            return default


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import re
import sqlite3
import typing

__all__ = ["SqliteSQL"]


class SqliteSQL:
    """Sqlite SQL generator"""

    @staticmethod
    def title_to_snake_case(string: str):
        """Title string to snake case
        >>> SqliteSQL.title_to_snake_case("HelloWorld")
        'hello_world'
        >>> SqliteSQL.title_to_snake_case("HelloHTMLWorldPeace")
        'hello_html_world_peace'
        """
        return re.sub(
            r"([A-Z]+)([A-Z][a-z])",
            r"\1_\2",
            re.sub(r"([a-z\d])([A-Z])", r"\1_\2", string),
        ).lower()

    @classmethod
    def clear_free_pages(cls, db: sqlite3.Connection, tries=1, chunk_size=100):
        db.execute("PRAGMA auto_vacuum = INCREMENTAL;")
        sql = "PRAGMA incremental_vacuum({chunk_size})".format(chunk_size=chunk_size)
        for _ in range(tries):
            db.execute(sql)
            db.commit()
        new_count = db.execute("PRAGMA freelist_count").fetchone()[0]
        return new_count

    @classmethod
    def create_table(
        cls,
        table: str,
        data_types: typing.Dict[str, typing.Type],
        primary_key: typing.Union[tuple, str] = (),
        unique_indexes: typing.Optional[list] = None,
        indexes: typing.Optional[list] = None,
        autoincrement=False,
        strict=True,
    ) -> list:
        r"""

        >>> from dataclasses import dataclass
        >>> @dataclass
        ... class User:
        ...     id: int
        ...     name: str
        ...     age: int
        ...     score: float
        ...     image: bytes
        ...     primary_key = "id"
        ...     autoincrement = True
        ...     indexes = [["age", "score"], "score"]
        ...     unique_indexes = ["name", "image"]
        >>> SqliteSQL.create_table("user", User.__annotations__)
        ['CREATE TABLE IF NOT EXISTS user(`id` INTEGER,`name` TEXT,`age` INTEGER,`score` REAL,`image` BLOB);']
        >>> SqliteSQL.create_table("user", User.__annotations__, primary_key=User.primary_key, unique_indexes=User.unique_indexes, indexes=User.indexes, autoincrement=User.autoincrement)
        ['CREATE TABLE IF NOT EXISTS user(`id` INTEGER,`name` TEXT,`age` INTEGER,`score` REAL,`image` BLOB, PRIMARY KEY (id AUTOINCREMENT));', 'CREATE UNIQUE INDEX IF NOT EXISTS `idx_name` ON user (`name`);', 'CREATE UNIQUE INDEX IF NOT EXISTS `idx_image` ON user (`image`);', 'CREATE INDEX IF NOT EXISTS `idx_age_score` ON user (`age`, `score`);', 'CREATE INDEX IF NOT EXISTS `idx_score` ON user (`score`);']
        >>> from typing import TypedDict
        >>> class User2(TypedDict):
        ...     id: int
        ...     name: str
        >>> SqliteSQL.create_table("user2", User2.__annotations__)
        ['CREATE TABLE IF NOT EXISTS user2(`id` INTEGER,`name` TEXT);']
        >>> SqliteSQL.create_table("user2", User2.__annotations__, primary_key=("id", "name"), unique_indexes=["name"])
        ['CREATE TABLE IF NOT EXISTS user2(`id` INTEGER,`name` TEXT, PRIMARY KEY (id, name));', 'CREATE UNIQUE INDEX IF NOT EXISTS `idx_name` ON user2 (`name`);']
        """
        type_mappings = {str: "TEXT", int: "INTEGER", float: "REAL", bytes: "BLOB"}
        table_sql = f"CREATE TABLE IF NOT EXISTS {table}"
        table_sql += "("
        if strict:
            not_supported = set(data_types.values()) - set(type_mappings.keys())
            if not_supported:
                raise TypeError(f"Unsupported types: {not_supported}")
        for key, value in data_types.items():
            tp = type_mappings.get(value, "TEXT")
            table_sql += f"`{key}` {tp},"
        if primary_key:
            if isinstance(primary_key, str):
                primary_key = (primary_key,)
            pk = ", ".join(primary_key)
            if autoincrement:
                pk += " AUTOINCREMENT"
            table_sql += f" PRIMARY KEY ({pk}),"

        table_sql = table_sql.rstrip(", ")
        table_sql += ");"
        result = [table_sql]
        if unique_indexes:
            for index in unique_indexes:
                result.append(cls.create_index(table, index, unique=True))
        if indexes:
            for index in indexes:
                result.append(cls.create_index(table, index, unique=False))
        return result

    @classmethod
    def create_index(
        cls, table: str, index: typing.Union[str, typing.Sequence[str]], unique=False
    ):
        r"""
        >>> SqliteSQL.create_index("user", "name")
        'CREATE INDEX IF NOT EXISTS `idx_name` ON user (`name`);'
        >>> SqliteSQL.create_index("user", ["name", "age"])
        'CREATE INDEX IF NOT EXISTS `idx_name_age` ON user (`name`, `age`);'
        >>> SqliteSQL.create_index("user", ["name", "age"], unique=True)
        'CREATE UNIQUE INDEX IF NOT EXISTS `idx_name_age` ON user (`name`, `age`);'
        """
        if isinstance(index, str):
            index = (index,)
        idx_name = "idx_" + "_".join(index)
        idx_value = ", ".join([f"`{i}`" for i in index])
        uniq = "UNIQUE " if unique else ""
        return (
            f"CREATE {uniq}INDEX IF NOT EXISTS `{idx_name}` ON {table} ({idx_value});"
        )

    @classmethod
    def insert(
        cls,
        table: str,
        keys: typing.Sequence,
        method: typing.Literal["insert", "replace"],
    ):
        """Generate insert sql.

        >>> user = dict(id=1, name="name", age=1, score=1.0, image=b"image")
        >>> SqliteSQL.insert("user", user.keys(), method="insert")
        'INSERT INTO user (`id`, `name`, `age`, `score`, `image`) VALUES (:id, :name, :age, :score, :image);'
        >>> SqliteSQL.insert("user", user.keys(), method="replace")
        'REPLACE INTO user (`id`, `name`, `age`, `score`, `image`) VALUES (:id, :name, :age, :score, :image);'
        """
        _keys = ", ".join([f"`{k}`" for k in keys])
        _values = ", ".join([f":{k}" for k in keys])
        return f"{method.upper()} INTO {table} ({_keys}) VALUES ({_values});"

    @classmethod
    def upsert(
        cls,
        table: str,
        keys: typing.Sequence,
        primary_key: typing.Union[str, typing.Sequence[str]],
    ):
        """Generate upsert(insert or update) sql.

        >>> user = dict(id=1, name="name", age=1, score=1.0, image=b"image")
        >>> SqliteSQL.upsert("user", user.keys(), primary_key="id")
        'INSERT OR REPLACE INTO user (`id`, `name`, `age`, `score`, `image`) VALUES (:id, :name, :age, :score, :image) ON CONFLICT(`id`) DO UPDATE SET (`id`, `name`, `age`, `score`, `image`) = (:id, :name, :age, :score, :image);'
        >>> SqliteSQL.upsert("user", user.keys(), primary_key=("id", "name"))
        'INSERT OR REPLACE INTO user (`id`, `name`, `age`, `score`, `image`) VALUES (:id, :name, :age, :score, :image) ON CONFLICT(`id`, `name`) DO UPDATE SET (`id`, `name`, `age`, `score`, `image`) = (:id, :name, :age, :score, :image);'
        """
        _keys = ", ".join([f"`{k}`" for k in keys])
        _values = ", ".join([f":{k}" for k in keys])
        if isinstance(primary_key, str):
            primary_key = (primary_key,)
        pk = ", ".join([f"`{k}`" for k in primary_key])
        return f"INSERT OR REPLACE INTO {table} ({_keys}) VALUES ({_values}) ON CONFLICT({pk}) DO UPDATE SET ({_keys}) = ({_values});"

    @classmethod
    def update(cls, table: str, update_keys: typing.Sequence[str], where: str = ""):
        """Generate update sql.

        >>> user = dict(id=1, name="name", age=1, score=1.0, image=b"image")
        >>> SqliteSQL.update("user", user.keys(), where="")
        'UPDATE user SET (`id`, `name`, `age`, `score`, `image`) = (:id, :name, :age, :score, :image);'
        >>> SqliteSQL.update("user", ("name", "score"), where="`id` = 10")
        'UPDATE user SET (`name`, `score`) = (:name, :score) WHERE `id` = 10;'
        """
        _keys = ", ".join([f"`{k}`" for k in update_keys])
        _values = ", ".join([f":{k}" for k in update_keys])
        result = f"UPDATE {table} SET ({_keys}) = ({_values})"
        if where:
            result += f" WHERE {where}"
        result += ";"
        return result

    @classmethod
    def delete(cls, table: str, where: str = "", limit=0):
        r"""Generate delete sql.

        >>> SqliteSQL.delete("user")
        'DELETE FROM user;'
        >>> SqliteSQL.delete("user", limit=100)
        'DELETE FROM user LIMIT 100;'
        >>> SqliteSQL.delete("user", where="`id` < 10")
        'DELETE FROM user WHERE `id` < 10;'
        """
        result = f"DELETE FROM {table}"
        if where:
            result += f" WHERE {where}"
        if limit:
            result += f" LIMIT {limit}"
        result += ";"
        return result


# class Q:
#     """[WIP]
#     Where clause generator.
#     Q("name").IN("John", "Tom", "Jerry").AND("age").BETWEEN(20, 30).OR("city").IN("New York", "Los Angeles").str()
#     # name IN ('John', 'Tom', 'Jerry') AND age BETWEEN 20 AND 30 OR city IN ('New York', 'Los Angeles')
#     """


# class KVSqlite:
#     """[WIP]
#     Sqlite KV store. Production environment recommends using `sqlitedict`."""


if __name__ == "__main__":
    from doctest import testmod

    testmod()
